import time
from typing import Any, Dict

from glQiwiApi.core.cache.constants import ADD_TIME_PLACEHOLDER, VALUE_PLACEHOLDER


def embed_cache_time(**kwargs: Any) -> Dict[Any, Any]:
    return {
        key: {VALUE_PLACEHOLDER: value, ADD_TIME_PLACEHOLDER: time.monotonic()}
        for key, value in kwargs.items()
    }



from typing import ClassVar, NoReturn, Optional

from pydantic import BaseModel


class YooMoneyErrorSchema(BaseModel):
    error_code: str
    error_description: Optional[str] = None
    status: Optional[str] = None


class YooMoneyError(Exception):
    """
    There are some facts about YooMoney API errors:

    * Error in json format contains at least one field - error.
    * Error field defines short description of error(like error code).

    * Sometimes error can contain field error_description,
      that enrich error code with some human-readable explanation.
      Also, error can contain status field and some technical fields.

    """

    match_error_code: ClassVar[Optional[str]] = None
    _explanation: ClassVar[Optional[str]] = None

    def __init__(self, error_schema: YooMoneyErrorSchema) -> None:
        self.error_schema = error_schema

        if error_schema.error_description is None and self._explanation is not None:
            error_schema.error_description = self._explanation

    def __str__(self) -> str:
        str_ = ''
        if self.error_schema.error_description:
            str_ += f'{self.error_schema.error_description}\n'
            str_ += f'For debugging purposes: \n'
            str_ += f"{'':>4}error code = {self.error_schema.error_code}"
            if self.error_schema.status:
                str_ += f"{'':>4}status={self.error_schema.status}"
        else:
            str_ += f'{self.error_schema.error_code}'
            if self.error_schema.status:
                str_ += ','
                str_ += f'status={self.error_schema.status}'

        return str_

    def json(self) -> str:
        return self.error_schema.json()

    @classmethod
    def raise_most_appropriate_error(cls, error_schema: YooMoneyErrorSchema) -> NoReturn:
        for subclass in cls.__subclasses__():
            if not subclass.is_i_matching_error_code(error_schema):
                continue

            raise subclass(error_schema)

        raise TechnicalError(error_schema)

    @classmethod
    def is_i_matching_error_code(cls, error_schema: YooMoneyErrorSchema) -> bool:
        return error_schema.error_code == cls.match_error_code


class InvalidRequestError(YooMoneyError):
    match_error_code = 'invalid_request'
    _explanation = 'The request is missing required parameters, or parameters have unsupported or invalid values.'


class UnauthorizedClientError(YooMoneyError):
    match_error_code = 'unauthorized_client'
    _explanation = 'The client_id value is invalid, or the application does not have rights to request authorization (for example, its client_id has been blocked by YooMoney).'


class InvalidGrantError(YooMoneyError):
    match_error_code = 'invalid_grant'
    _explanation = 'The access_token could not be issued. Either the temporary authorization code was not issued by YooMoney, or it has expired, or an access_token has already been issued for this temporary authorization code (a duplicate request for an access token using the same temporary authorization code).'


class InsufficientScopeError(YooMoneyError):
    match_error_code = 'insufficient_scope'
    _explanation = 'The token does not have permissions for the requested operation.'


class IllegalParamError(YooMoneyError):
    def __init__(self, error_schema: YooMoneyErrorSchema):
        super(IllegalParamError, self).__init__(error_schema)
        _, _, self.param_name = self.error_schema.error_code.rpartition('illegal_param_')
        self.error_schema.error_description = f'Illegal value of "{self.param_name}" param.'

    @classmethod
    def is_i_matching_error_code(cls, error_schema: YooMoneyErrorSchema) -> bool:
        if not error_schema.error_code.startswith('illegal_param_'):
            return False

        return True


class IllegalParamsError(YooMoneyError):
    match_error_code = 'illegal_params'
    _explanation = 'Required payment parameters are either missing or have invalid values.'


class NotEnoughFundsError(YooMoneyError):
    match_error_code = 'not_enough_funds'
    _explanation = 'The payer’s account does not have sufficient funds to make the payment. Additional funds should be credited to the account, and a new payment will need to be processed.'


class PayeeNotFoundError(YooMoneyError):
    match_error_code = 'payee_not_found'
    _explanation = 'The transfer recipient was not found. The specified account does not exist, or a phone number or email address was specified that is not linked to a user account or payee.'


class PaymentRefusedError(YooMoneyError):
    match_error_code = 'payment_refused'
    _explanation = 'The merchant refused to accept the payment (for example, the user tried to purchase an item that is not in stock).'


class AuthorizationRejectError(YooMoneyError):
    match_error_code = 'authorization_reject'
    _explanation = """Authorization of the payment was refused. Possible reasons:
A transaction with the current parameters is forbidden for this user.
The user did not accept the User Agreement for the YooMoney service.
"""


class LimitExceededError(YooMoneyError):
    match_error_code = 'limit_exceeded'
    _explanation = """One of the operation limits was exceeded:
For the total amount of operations for the access token granted.
For the total amount of operations over a period of time for the access token granted.
YooMoney restrictions for various types of operations."""


class AccountBlockedError(YooMoneyError):
    match_error_code = 'account_blocked'
    _explanation = 'The user’s account has been blocked. In order to unblock the account, the user must be redirected to the address specified in the account_unblock_uri field.'


class AccountClosedError(YooMoneyError):
    match_error_code = 'account_closed'
    _explanation = 'User’s account closed.'


class ExtActionRequiredError(YooMoneyError):
    match_error_code = 'ext_action_required'
    _explanation = """This type of payment cannot be made at this time.
To be able to make these types of payments, the user must go to the page with the ext_action_uri address and follow the instructions on that page.
This may be any of the following actions:
    * Entering identification data.
    * Accepting the offer.
    * Performing other actions according to the instructions."""


class ContractNotFoundError(YooMoneyError):
    match_error_code = 'contract_not_found'
    _explanation = 'There is no existing unconfirmed payment with the specified request_id.'


class MoneySourceNotAvailableError(YooMoneyError):
    match_error_code = 'money_source_not_available'
    _explanation = 'The requested payment method (money_source) is not available for this payment.'


class TechnicalError(YooMoneyError):
    """Acting like a fallback error."""



from __future__ import annotations

import http
from dataclasses import dataclass
from typing import Any, Dict, Optional, Tuple, Union


class Payload:
    def __init__(
        self,
        headers: Optional[Dict[Any, Any]] = None,
        json: Optional[Dict[Any, Any]] = None,
        params: Optional[Dict[Any, Any]] = None,
        data: Optional[Dict[Any, Any]] = None,
        **kwargs: Any,
    ) -> None:
        self.headers = headers
        self.json = json
        self.params = params
        self.data = data

    @classmethod
    def new(cls, kwargs: Dict[Any, Any], args: Tuple[Any, ...]) -> Payload:
        return cls(**{k: kwargs.get(k) for k in args if isinstance(kwargs.get(k), dict)})


@dataclass(frozen=True)
class CachedAPIRequest:
    payload: Payload
    response: Any
    method: Union[str, http.HTTPStatus]



from __future__ import annotations

import abc
import inspect
from typing import Any, Awaitable, Callable, Generic, Tuple, Type, TypeVar, Union, cast

Event = TypeVar('Event')


class BaseFilter(abc.ABC, Generic[Event]):
    @abc.abstractmethod
    async def check(self, update: Event) -> bool:
        raise NotImplementedError

    def __and__(self, other: BaseFilter[Event]) -> AndFilter[Event]:
        if not isinstance(other, BaseFilter):
            raise TypeError(
                f"Can't compose two different types of filters, expected Filter, got {type(other)}"
            )
        return AndFilter(self, other)

    def __invert__(self) -> NotFilter[Event]:
        return NotFilter(self)


class AndFilter(BaseFilter[Event]):
    def __init__(self, filter1: BaseFilter[Event], filter2: BaseFilter[Event]) -> None:
        self.filter2 = filter2
        self.filter1 = filter1

    async def check(self, update: Any) -> bool:
        return await self.filter1.check(update) and await self.filter2.check(update)


class NotFilter(BaseFilter[Event]):
    def __init__(self, filter_: BaseFilter[Event]):
        self.filter_ = filter_

    async def check(self, update: Event) -> bool:
        return not await self.filter_.check(update)


class LambdaBasedFilter(BaseFilter[Event]):
    def __init__(self, func: Callable[[Event], Union[bool, Awaitable[bool]]]) -> None:
        self.name = f'Filter around <{func!r}>'

        self.function = func
        self.awaitable: bool = inspect.iscoroutinefunction(func) or inspect.isawaitable(func)

    async def check(self, update: Event) -> bool:
        if self.awaitable:
            return await cast(Awaitable[bool], self.function(update))
        else:
            return cast(bool, self.function(update))


class ExceptionFilter(BaseFilter[Event]):
    def __init__(self, exception: Union[Type[Exception], Tuple[Type[Exception]]]):
        self._exception = exception

    async def check(self, update: Event) -> bool:
        return isinstance(update, self._exception)


__all__ = ('LambdaBasedFilter', 'BaseFilter', 'NotFilter', 'AndFilter', 'Event', 'ExceptionFilter')



import time
from typing import Any, ClassVar, Dict, Optional

from pydantic import Field, validator

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import PaymentInfo


class TransferMoney(QiwiAPIMethod[PaymentInfo]):
    url: ClassVar[str] = 'https://edge.qiwi.com/sinap/api/v2/terms/99/payments'
    http_method: ClassVar[str] = 'POST'

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        'id': RuntimeValue(default_factory=lambda: str(int(time.time() * 1000))),
        'sum': {'amount': RuntimeValue(), 'currency': '643'},
        'paymentMethod': {'type': 'Account', 'accountId': '643'},
        'comment': RuntimeValue(mandatory=False),
        'fields': {'account': RuntimeValue()},
    }

    @validator('to_wallet')
    def add_plus_sign_to_phone_number(cls, v: str) -> str:
        if v.startswith('+'):
            return v
        return f'+{v}'

    amount: float = Field(..., scheme_path='sum.amount')
    to_wallet: str = Field(..., scheme_path='fields.account')
    comment: Optional[str] = Field(None, scheme_path='comment')



from __future__ import annotations

import asyncio
import functools
import inspect
import warnings
from typing import Any, Callable, Generic, Optional, Type, TypeVar, Union, cast

_T = TypeVar('_T')
_F = TypeVar('_F')
DeprecatedFunc = Union[Callable[..., _F], Callable[[Callable[..., _F]], _F]]


def deprecated(reason: Union[str, Callable[..., _F]], stacklevel: int = 2) -> DeprecatedFunc:
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    Source: https://stackoverflow.com/questions/2536307/decorators-in-the-python-standard-lib-deprecated-specifically
    """

    if isinstance(reason, str):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func: Callable[..., _F]) -> Callable[..., _F]:

            if inspect.isclass(func):
                msg = 'Call to deprecated class {name} ({reason}).'
            else:
                msg = 'Call to deprecated function {name} ({reason}).'

            @functools.wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> _F:
                warn_deprecated(
                    msg.format(name=func.__name__, reason=reason), stacklevel=stacklevel
                )
                warnings.simplefilter('default', DeprecationWarning)
                return func(*args, **kwargs)

            return wrapper

        return decorator

    if inspect.isclass(reason) or inspect.isfunction(reason):

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func1 = reason  # type: ignore

        if inspect.isclass(func1):
            msg1 = 'Call to deprecated class {name}.'
        else:
            msg1 = 'Call to deprecated function {name}.'

        @functools.wraps(func1)
        def wrapper1(*args: Any, **kwargs: Any) -> _F:
            warn_deprecated(msg1.format(name=func1.__name__), stacklevel=stacklevel)  # type: ignore
            return cast(_F, func1(*args, **kwargs))  # type: ignore

        return wrapper1

    raise TypeError(repr(type(reason)))


def warn_deprecated(message: str, warning: Any = DeprecationWarning, stacklevel: int = 2) -> None:
    warnings.simplefilter('always', warning)
    warnings.warn(message, category=warning, stacklevel=stacklevel)
    warnings.simplefilter('default', warning)


def renamed_argument(
    old_name: str, new_name: str, until_version: str, stacklevel: int = 3
) -> Callable[[Callable[..., _F]], _F]:
    """
    A meta-decorator to mark an argument as deprecated.
    .. code-block:: python3
        @renamed_argument("chat", "chat_id", "3.0")  # stacklevel=3 by default
        @renamed_argument("user", "user_id", "3.0", stacklevel=4)
        def some_function(user_id, chat_id=None):
            print(f"user_id={user_id}, chat_id={chat_id}")
        some_function(user=123)  #  prints 'user_id=123, chat_id=None' with warning
        some_function(123)  #  prints 'user_id=123, chat_id=None' without warning
        some_function(user_id=123)  #  prints 'user_id=123, chat_id=None' without warning
    :param old_name:
    :param new_name:
    :param until_version: the version in which the argument is scheduled to be removed
    :param stacklevel: leave it to default if it's the first decorator used.
    Increment with any new decorator used.
    :return: decorator
    """

    def decorator(func):
        is_coroutine = asyncio.iscoroutinefunction(func)

        def _handling(kwargs):
            """
            Returns updated version of kwargs.
            """
            routine_type = 'coroutine' if is_coroutine else 'function'
            if old_name in kwargs:
                warn_deprecated(
                    f"In {routine_type} '{func.__name__}' argument '{old_name}' "
                    f"is renamed to '{new_name}' "
                    f'and will be removed in aiogram {until_version}',
                    stacklevel=stacklevel,
                )
                kwargs = kwargs.copy()
                kwargs.update({new_name: kwargs.pop(old_name)})
            return kwargs

        if is_coroutine:

            @functools.wraps(func)
            async def wrapped(*args, **kwargs) -> _F:
                kwargs = _handling(kwargs)
                return await func(*args, **kwargs)

        else:

            @functools.wraps(func)
            def wrapped(*args, **kwargs) -> _F:
                kwargs = _handling(kwargs)
                return func(*args, **kwargs)

        return wrapped

    return decorator


_VT = TypeVar('_VT')
_OwnerCls = TypeVar('_OwnerCls')


class DeprecatedReadOnlyClassVar(Generic[_OwnerCls, _VT]):
    """
    DeprecatedReadOnlyClassVar[Owner, ValueType]
    :param warning_message: Warning message when getter gets called
    :param new_value_getter: Any callable with (owner_class: Type[Owner]) -> ValueType
                             signature that will be executed
    Usage example:
    >>> class MyClass:
    ...     some_attribute: DeprecatedReadOnlyClassVar[MyClass, int] = \
    ...            DeprecatedReadOnlyClassVar(
    ...                  "Warning message.", lambda owner: 15)
    ...
    >>> MyClass.some_attribute  # does warning.warn with `Warning message` and returns 15 in the current case
    """

    __slots__ = '_new_value_getter', '_warning_message'

    def __init__(self, warning_message: str, new_value_getter: Callable[[Type[_OwnerCls]], _VT]):
        self._warning_message = warning_message
        self._new_value_getter = new_value_getter

    def __get__(self, instance: Optional[_OwnerCls], owner: Type[_OwnerCls]) -> _VT:
        warn_deprecated(self._warning_message, stacklevel=3)
        return self._new_value_getter(owner)



from typing import Any, Dict, Optional

from pydantic import Field

from glQiwiApi.types.amount import AmountWithCurrency
from glQiwiApi.types.base import Base


class Fields(Base):
    """object: Fields"""

    account: str


class State(Base):
    """object: State"""

    code: str


class TransactionInfo(Base):
    """object: TransactionInfo"""

    id: int
    state: State


class PaymentInfo(Base):
    """object: PaymentInfo"""

    id: int
    amount: AmountWithCurrency = Field(..., alias='sum')
    terms: str
    fields: Fields
    source: str
    transaction: Optional[TransactionInfo] = None
    comment: Optional[str] = None


class PaymentMethod(Base):
    type: str = 'Account'
    account_id: int = Field(643, alias='accountId')


class QiwiPayment(Base):
    id: int
    sum: AmountWithCurrency
    method: PaymentMethod = Field(..., alias='paymentMethod')
    fields: Dict[Any, Any]
    comment: Optional[str] = None


__all__ = ['PaymentInfo', 'QiwiPayment']



from __future__ import annotations

import re
from collections import namedtuple
from typing import Any, Optional, Pattern, Type, TypeVar, cast

from glQiwiApi.core.event_fetching.webhooks.config import DEFAULT_QIWI_WEBHOOK_PATH

_URL = TypeVar('_URL', bound='WebhookURL')

# Url without port e.g. https://127.0.0.1/ or https://website.com/
HOST_REGEX: Pattern[str] = re.compile(
    r'^(http(s?)://)?'
    r'(((www\.)?[a-zA-Z0-9.\-_]+'
    r'(\.[a-zA-Z]{2,6})+)|'
    r'(\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.)'
    r'{3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\b))'
    r'(/[a-zA-Z0-9_\-\s./?%#&=]*)?$'
)


class WebhookURL(
    namedtuple(
        'WebhookURL',
        ['host', 'webhook_path', 'port', 'https'],
    )
):
    host: str
    webhook_path: Optional[str]
    port: Optional[int]
    https: bool = False

    @classmethod
    def create(
        cls: Type[_URL],
        host: str,
        port: Optional[int] = None,
        webhook_path: Optional[str] = None,
        https: bool = False,
    ) -> _URL:
        return cls(
            host=cls._assert_host(host, param_name='host'),
            webhook_path=cls._assert_str(webhook_path, param_name='webhook_path'),
            port=cls._assert_int(port, param_name='port'),
            https=https,
        )

    @classmethod
    def _assert_int(cls, v: Optional[Any], *, param_name: str) -> Optional[int]:
        if v is None:
            return v

        if not isinstance(v, int):
            raise TypeError('%s must be integer' % param_name)
        return v

    @classmethod
    def _assert_str(cls, v: Optional[Any], *, param_name: str) -> Optional[str]:
        if v is None:
            return v

        if not isinstance(v, str):
            raise TypeError('%s must be a string' % param_name)
        return v

    @classmethod
    def _assert_host(cls, v: Any, *, param_name: str) -> str:
        if not re.match(HOST_REGEX, v):
            raise TypeError(
                '%s must be like https://127.0.0.1/ or https://website.com/' % param_name
            )
        return cast(str, v)

    def render(self) -> str:
        host = self.host
        if self.webhook_path is None:
            # Here we use `DEFAULT_QIWI_WEBHOOK_PATH` instead of DEFAULT_QIWI_BILLS_WEBHOOK_PATH
            # because the second you need to register directly in QIWI P2P API and it's no need to build endpoint to it
            self.webhook_path = DEFAULT_QIWI_WEBHOOK_PATH
        if self.port is not None:
            host += f':{self.port}'
        if self._doesnt_contains_slash():
            host += '/'
        if self.https:
            scheme = 'https://'
        else:
            scheme = 'http://'
        return f'{scheme}{host}{self.webhook_path}'

    def _doesnt_contains_slash(self) -> bool:
        return not (self.host.endswith('/') and self.webhook_path.startswith('/'))  # type: ignore



import logging
from typing import Any, Awaitable, Callable

from aiohttp import web
from aiohttp.typedefs import Handler
from aiohttp.web_middlewares import middleware

from glQiwiApi.core.event_fetching.webhooks.services.security.ip import IPFilter
from glQiwiApi.core.event_fetching.webhooks.utils import check_ip

logger = logging.getLogger('glQiwiApi.webhooks.middlewares')


def ip_filter_middleware(ip_filter: IPFilter) -> Callable[[web.Request, Handler], Awaitable[Any]]:
    @middleware
    async def _ip_filter_middleware(request: web.Request, handler: Handler) -> Any:
        ip_address, accept = check_ip(ip_filter=ip_filter, request=request)
        if not accept:
            logger.warning(f'Blocking request from an unauthorized IP: {ip_address}')
            raise web.HTTPUnauthorized()
        return await handler(request)

    return _ip_filter_middleware



from typing import ClassVar, Dict

from glQiwiApi.core.abc.api_method import APIMethod


class RejectIncomingTransfer(APIMethod[Dict[str, str]]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://yoomoney.ru/api/incoming-transfer-reject'

    operation_id: str



from typing import ClassVar, Union

from glQiwiApi.core.abc.api_method import APIMethod
from glQiwiApi.yoo_money.types import OperationDetails


class OperationDetailsMethod(APIMethod[OperationDetails]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://yoomoney.ru/api/operation-details'

    operation_id: Union[int, str]



import sys
from typing import Any, AnyStr, AsyncContextManager

if sys.version_info >= (3, 9):

    def remove_suffix(input_string: AnyStr, suffix: AnyStr) -> AnyStr:
        return input_string.removesuffix(suffix)

else:

    def remove_suffix(input_string: AnyStr, suffix: AnyStr) -> AnyStr:
        """Backport for python 3.9 str.removesuffix(...)"""
        if suffix and input_string.endswith(suffix):
            return input_string[: -len(suffix)]
        return input_string


class ModuleNotInstalledException(Exception):
    pass


class EmptyCls(object):
    def __init__(self) -> None:
        raise ModuleNotInstalledException()


try:
    import orjson as json
except ImportError:
    import json

try:
    import aiofiles
except ImportError:

    class aiofiles_compat:
        def open(self, *args: Any, **kwargs: Any) -> AsyncContextManager[Any]:
            raise ModuleNotInstalledException(
                "Module aiofiles not installed and you can't use it's "
                'functionality till you install this module.'
            )

    aiofiles = aiofiles_compat()  # type: ignore

try:
    import orjson as json  # noqa
except (ImportError, ModuleNotFoundError):
    import json  # type: ignore

try:
    from aiogram import Dispatcher
    from aiogram.types import InputFile
except (ModuleNotFoundError, ImportError):
    Dispatcher = EmptyCls
    InputFile = EmptyCls

if sys.version_info >= (3, 8):
    from typing import Final, Literal, Protocol  # noqa
else:
    from typing_extensions import Final, Literal, Protocol  # noqa



from __future__ import annotations

import abc
import asyncio
import inspect
import logging
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Awaitable, Callable, Dict, Iterable, List, Optional, Union, cast

try:
    import zoneinfo
except ImportError:
    import backports.zoneinfo as zoneinfo

from aiohttp import web
from aiohttp.web import _run_app  # noqa

from glQiwiApi import QiwiWrapper
from glQiwiApi.core.event_fetching.dispatcher import BaseDispatcher
from glQiwiApi.core.event_fetching.webhooks.app import configure_app
from glQiwiApi.core.event_fetching.webhooks.config import WebhookConfig
from glQiwiApi.ext.webhook_url import WebhookURL
from glQiwiApi.plugins.abc import Pluggable
from glQiwiApi.qiwi.clients.wallet.client import QiwiWallet
from glQiwiApi.qiwi.clients.wallet.methods.history import MAX_HISTORY_LIMIT
from glQiwiApi.qiwi.clients.wallet.types import History
from glQiwiApi.utils.synchronous import adapter

logger = logging.getLogger('glQiwiApi.executor')

TIMEOUT_IF_EXCEPTION = 40
DEFAULT_TIMEOUT = 5
WALLET_CTX_KEY = 'wallet'


class _NoUpdatesToExecute(Exception):
    """Internal exception to determine that history is empty"""


_EventHandlerType = Callable[['Context'], Union[None, Awaitable[None]]]


@dataclass
class _HandlerSpec:
    handler_fn: Callable[..., Any]

    @property
    def is_awaitable(self) -> bool:
        return inspect.isawaitable(self.handler_fn) or inspect.iscoroutinefunction(self.handler_fn)


class ExecutorEvent:
    def __init__(
        self,
        context: Context,
        init_handlers: Iterable[Optional[_EventHandlerType]] = (),
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        self._handlers: List[_HandlerSpec] = []
        for handler in init_handlers:
            if handler is None:
                continue
            self._handlers.append(_HandlerSpec(handler))
        self.context = context
        self._loop: asyncio.AbstractEventLoop = loop or asyncio.get_event_loop()

    def __iadd__(self, handler: Callable[..., Any]) -> 'ExecutorEvent':
        self._handlers.append(_HandlerSpec(handler))
        return self

    def __isub__(self, handler: Callable[..., Any]) -> 'ExecutorEvent':
        self._handlers.remove(_HandlerSpec(handler))
        return self

    def __len__(self) -> int:
        return len(self._handlers)

    async def fire(self) -> None:
        for handler_spec in self._handlers:
            if handler_spec.is_awaitable:
                return await handler_spec.handler_fn(self.context)  # type: ignore
            await self._loop.run_in_executor(None, handler_spec.handler_fn, self.context)


def start_webhook(
    wallet: Union[QiwiWallet, QiwiWrapper],
    dispatcher: BaseDispatcher,
    *plugins: Pluggable,
    webhook_config: WebhookConfig,
    on_startup: Optional[_EventHandlerType] = None,
    on_shutdown: Optional[_EventHandlerType] = None,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    context: Union[Dict[str, Any], Context, None] = None,
) -> None:
    """
    Blocking function that listens for webhooks.
    Supports only `glQiwiApi.types.BillWebhook` and `glQiwiApi.types.TransactionWebhook`

    :param wallet: instance of QiwiWrapper
    :param dispatcher:
    :param on_startup: coroutine,which will be executed on startup
    :param on_shutdown: coroutine, which will be executed on shutdown
    :param webhook_config:
    :param plugins: List of plugins, that will be executed together with polling.
         For example  builtin TelegramWebhookPlugin or other
         class, that implement Pluggable abc interface, deal with foreign framework/application
         in the background
    :param loop:
    :param context: context, that could be transmitted to handlers
    """
    if context is None:
        context = {}
    executor = WebhookExecutor(
        wallet,
        dispatcher,
        *plugins,
        on_shutdown=on_shutdown,
        on_startup=on_startup,
        loop=loop,
        context=Context(context),
    )
    executor.start_webhook(config=webhook_config)


def start_polling(
    wallet: Union[QiwiWallet, QiwiWrapper],
    dispatcher: BaseDispatcher,
    *plugins: Pluggable,
    skip_updates: bool = False,
    timeout_in_seconds: float = DEFAULT_TIMEOUT,
    on_startup: Optional[_EventHandlerType] = None,
    on_shutdown: Optional[_EventHandlerType] = None,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    context: Union[Dict[str, Any], Context, None] = None,
) -> None:
    """
    Setup for long-polling mode. Support only `glQiwiApi.types.Transaction` as event.

    :param wallet: instance of QiwiWrapper
    :param dispatcher:
    :param skip_updates:
    :param timeout_in_seconds: timeout of polling in seconds, if the timeout is too small,
         the API can throw an exception
    :param on_startup: function or coroutine,
         which will be executed on startup
    :param on_shutdown: function or coroutine,
         which will be executed on shutdown
    :param plugins: List of plugins, that will be executed together with polling.
         For example  builtin TelegramPollingPlugin or other
         class, that implement Pluggable abc interface, deal with foreign framework/application
         in the background
    :param loop:
    :param context: context, that could be transmitted to handlers
    """
    if context is None:
        context = {}
    executor = PollingExecutor(
        wallet,
        dispatcher,
        *plugins,
        timeout=timeout_in_seconds,
        skip_updates=skip_updates,
        on_startup=on_startup,
        on_shutdown=on_shutdown,
        loop=loop,
        context=Context(context),
    )
    executor.start_polling()


async def start_non_blocking_qiwi_api_polling(
    wallet: Union[QiwiWallet, QiwiWrapper],
    dispatcher: BaseDispatcher,
    skip_updates: bool = False,
    timeout_in_seconds: float = DEFAULT_TIMEOUT,
    on_startup: Optional[_EventHandlerType] = None,
    on_shutdown: Optional[_EventHandlerType] = None,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    context: Union[Dict[str, Any], Context, None] = None,
) -> asyncio.Task:
    if context is None:
        context = {}
    executor = PollingExecutor(
        wallet,
        dispatcher,
        timeout=timeout_in_seconds,
        skip_updates=skip_updates,
        on_startup=on_startup,
        on_shutdown=on_shutdown,
        loop=loop,
        context=Context(context),
    )
    return await executor.start_non_blocking_polling()


def configure_app_for_qiwi_webhooks(
    wallet: Union[QiwiWallet, QiwiWrapper],
    dispatcher: BaseDispatcher,
    app: web.Application,
    cfg: WebhookConfig,
) -> web.Application:
    executor = WebhookExecutor(wallet, dispatcher, context=Context({}))
    return executor.add_routes_for_webhook(app, cfg)


class Context(Dict[str, Any]):
    def __getattr__(self, item: str) -> Any:
        return self[item]

    @property
    def wallet(self) -> Union[QiwiWallet, QiwiWrapper]:
        return cast(Union[QiwiWallet, QiwiWrapper], self[WALLET_CTX_KEY])


class BaseExecutor(abc.ABC):
    def __init__(
        self,
        dispatcher: BaseDispatcher,
        *plugins: Pluggable,
        context: Context,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        on_startup: Optional[Callable[..., Any]] = None,
        on_shutdown: Optional[Callable[..., Any]] = None,
    ) -> None:
        if loop is not None:
            self._loop = loop  # pragma: no cover
        self._on_startup = ExecutorEvent(context, init_handlers=[on_startup], loop=self.loop)
        self._on_shutdown = ExecutorEvent(context, init_handlers=[on_shutdown], loop=self.loop)
        self._dispatcher = dispatcher
        self._plugins = plugins
        self._context = context

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        self._loop = cast(
            asyncio.AbstractEventLoop, getattr(self, '_loop', asyncio.get_event_loop())
        )
        # There is a issue, connected with aiohttp.web application, because glQiwiApi wants to have the same interface
        # for polling and webhooks on_startup & on_shutdown callbacks,
        # so we don't use app.on_shutdown.append(...), on the contrary we use self-written
        # system of callbacks, and so aiohttp.web.Application close event loop on shutdown, that's why we need to
        # create new event loop to gracefully execute shutdown callback.
        # It is highly undesirable to delete this line because you will always catch RuntimeWarning and RuntimeError.
        self._create_new_loop_if_closed()
        return self._loop

    def _create_new_loop_if_closed(self) -> None:
        if self._loop.is_closed():
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

    async def welcome(self) -> None:
        logger.info('Executor has started work!')
        await self._on_startup.fire()

    async def goodbye(self) -> None:
        logger.info('Goodbye!')
        await self._on_shutdown.fire()

    async def _install_plugins(self, ctx: Optional[Dict[Any, Any]] = None) -> None:
        if ctx is None:
            ctx = {}
        incline_tasks = [plugin.install(ctx) for plugin in self._plugins]
        await asyncio.gather(*incline_tasks)

    async def _shutdown(self) -> None:
        """
        On shutdown, executor gracefully cancel all tasks, close event loop
        and call `close` method to clear resources
        """
        callbacks = [self.goodbye(), self._shutdown_plugins()]
        await asyncio.shield(asyncio.gather(*callbacks))

    async def _shutdown_plugins(self) -> None:
        logger.debug('Shutting down plugins')
        shutdown_tasks = [asyncio.create_task(plugin.shutdown()) for plugin in self._plugins]
        await asyncio.gather(*shutdown_tasks)


class PollingExecutor(BaseExecutor):
    def __init__(
        self,
        wallet: Union[QiwiWallet, QiwiWrapper],
        dispatcher: BaseDispatcher,
        *plugins: Pluggable,
        context: Context,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        timeout: Union[float, int] = DEFAULT_TIMEOUT,
        skip_updates: bool = False,
        on_startup: Optional[_EventHandlerType] = None,
        on_shutdown: Optional[_EventHandlerType] = None,
    ) -> None:
        super(PollingExecutor, self).__init__(
            dispatcher,
            *plugins,
            loop=loop,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            context=context,
        )
        self.offset: Optional[int] = None
        self.get_updates_from = datetime.utcnow().astimezone(zoneinfo.ZoneInfo('Europe/Moscow'))
        self._timeout = _parse_timeout(timeout)
        self.skip_updates = skip_updates
        self._wallet = wallet

        self._context[WALLET_CTX_KEY] = self._wallet

    def start_polling(self) -> None:
        try:
            self.loop.create_task(self._install_plugins())
            self.loop.create_task(self._run_infinite_polling())
            adapter.run_forever_safe(self.loop)
        except (SystemExit, KeyboardInterrupt):  # pragma: no cover
            # allow graceful shutdown
            pass

    async def start_non_blocking_polling(self) -> asyncio.Task:
        return asyncio.create_task(self._run_infinite_polling())

    async def _run_infinite_polling(self) -> None:
        default_timeout = self._timeout

        try:
            await self.welcome()
            while True:
                try:
                    await self._try_fetch_new_updates()
                    self._set_timeout(default_timeout)
                except Exception as ex:
                    self._set_timeout(TIMEOUT_IF_EXCEPTION)
                    logger.error('Handle !r. Sleeping %s seconds', ex, self._timeout)
                await asyncio.sleep(self._timeout)
        finally:
            await asyncio.shield(asyncio.gather(self.goodbye(), self._wallet.close()))

    async def _try_fetch_new_updates(self) -> None:
        try:
            history = await self._fetch_history()
        except _NoUpdatesToExecute:
            return None
        if self.offset is None:
            first_update = history[0]
            self.offset = first_update.id - 1
        logger.debug('Current transaction offset is %d', self.offset)
        await self.process_updates(history)

    async def _fetch_history(self) -> History:
        end_date = datetime.utcnow().astimezone(zoneinfo.ZoneInfo('Europe/Moscow'))
        if isinstance(self._wallet, QiwiWallet):
            history = (
                await self._wallet.history(start_date=self.get_updates_from, end_date=end_date)
            ).sorted_by_date()
        else:
            history = (
                await self._wallet.transactions(
                    start_date=self.get_updates_from, end_date=end_date
                )
            ).sorted_by_date()

        if len(history) == MAX_HISTORY_LIMIT:
            logger.debug('History is out of max history transaction limit')
            first_txn_by_date = history[-1]
            self.get_updates_from = first_txn_by_date.date

        if self.skip_updates:
            self.skip_updates = False
            raise _NoUpdatesToExecute()
        elif not history:
            raise _NoUpdatesToExecute()

        return history

    async def process_updates(self, history: History) -> None:
        tasks: List[asyncio.Task[None]] = [
            asyncio.create_task(self._dispatcher.process_event(event, self._context))
            for event in history
            if cast(int, self.offset) < event.id
        ]
        if history:
            self.offset = history.sorted_by_id().last().id
        await asyncio.gather(*tasks)

    def _set_timeout(self, exception_timeout: Union[int, float]) -> None:
        self._timeout = exception_timeout

    async def _shutdown(self) -> None:
        await asyncio.gather(super()._shutdown(), self._wallet.close())


class WebhookExecutor(BaseExecutor):
    def __init__(
        self,
        wallet: Union[QiwiWallet, QiwiWrapper],
        dispatcher: BaseDispatcher,
        *plugins: Pluggable,
        context: Context,
        on_startup: Optional[_EventHandlerType] = None,
        on_shutdown: Optional[_EventHandlerType] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        super().__init__(
            dispatcher,
            *plugins,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            loop=loop,
            context=context,
        )
        self._application = web.Application(loop=self.loop)
        self._wallet = wallet

        self._context[WALLET_CTX_KEY] = self._wallet

    def add_routes_for_webhook(
        self, app: web.Application, config: WebhookConfig
    ) -> web.Application:
        supplemented_configuration = self.loop.run_until_complete(
            self._supplement_configuration(config)
        )
        try:
            return configure_app(
                dispatcher=self._dispatcher,
                app=app,
                webhook_config=supplemented_configuration,
            )
        finally:
            self.loop.run_until_complete(self._wallet.close())

    def start_webhook(self, config: WebhookConfig) -> None:
        supplemented_configuration = self.loop.run_until_complete(
            self._supplement_configuration(config)
        )

        self._application = configure_app(
            dispatcher=self._dispatcher,
            app=self._application,
            webhook_config=supplemented_configuration,
        )

        self.loop.create_task(self._install_plugins())
        try:
            self.loop.run_until_complete(self.welcome())
            self.loop.create_task(
                _run_app(
                    app=self._application,
                    host=supplemented_configuration.app.host,
                    port=supplemented_configuration.app.port,
                    ssl_context=supplemented_configuration.app.ssl_context,
                    **supplemented_configuration.app.kwargs,
                )
            )
            adapter.run_forever_safe(self.loop)
        finally:
            self.loop.run_until_complete(self.goodbye())

    async def _supplement_configuration(self, config: WebhookConfig) -> WebhookConfig:
        config = deepcopy(config)
        if config.app.base_app is not None:
            self._application = config.app.base_app

        if config.encryption.secret_p2p_key is None:
            raise RuntimeError(
                'Secret p2p token is empty, cannot setup webhook without it. '
                'Please, provide token to work with webhooks.'
            )

        await self._set_base64_encryption_key_for_config(config)
        return config

    async def _set_base64_encryption_key_for_config(self, config: WebhookConfig) -> None:
        if config.encryption.base64_encryption_key is None:
            if config.hook_registration.host_or_ip_address is None:
                raise RuntimeError(
                    "You didn't transmit neither base64 "
                    'encryption key to WebhookConfig nor host or ip address to bind new webhook.'
                    'To fix it, pass on WebhookConfig.hook_registration.host_or_ip_address or '
                    'WebhookConfig.encryption.base64_encryption_key to executor.start_webhook(...)'
                )
            _, base64_encryption_key = await self._wallet.bind_webhook(
                url=WebhookURL.create(
                    host=config.hook_registration.host_or_ip_address,
                    port=config.app.port,
                    webhook_path=config.routes.standard_qiwi_hook_path[1:],
                    https=config.app.ssl_context is not None,
                ),
                delete_old=True,
            )
            config.encryption.base64_encryption_key = base64_encryption_key


def _parse_timeout(timeout: Union[float, int]) -> float:  # pragma: no cover
    if isinstance(timeout, float):
        return timeout
    elif isinstance(timeout, int):
        return float(timeout)
    else:
        raise TypeError(f'Timeout must be float or int. You have passed on {type(timeout)}')



from __future__ import annotations

import abc
from typing import Any

from .base import Handler


class ErrorHandler(Handler[Exception], abc.ABC):
    def __init__(self, event: Exception, *args: Any) -> None:
        super().__init__(event)
        self.args = args

    @property
    def exception_name(self) -> str:
        return self.event.__class__.__qualname__



from __future__ import annotations

import abc
import time
from typing import TYPE_CHECKING, Any, Dict, Union

from glQiwiApi.core.cache.cached_types import CachedAPIRequest, Payload
from glQiwiApi.core.cache.constants import ADD_TIME_PLACEHOLDER
from glQiwiApi.core.cache.exceptions import CacheExpiredError, CacheValidationError

if TYPE_CHECKING:
    from glQiwiApi.core.cache.storage import CacheStorage

INFINITE = float('inf')


class CacheInvalidationStrategy(abc.ABC):
    @abc.abstractmethod
    async def process_update(self, **kwargs: Any) -> None:
        pass

    async def process_delete(self) -> None:
        pass

    @abc.abstractmethod
    async def process_retrieve(self, **kwargs: Any) -> None:
        pass

    async def check_is_contains_similar(self, cache_storage: CacheStorage, item: Any) -> bool:
        raise TypeError

    @property
    def is_cache_disabled(self) -> bool:
        return False


class UnrealizedCacheInvalidationStrategy(CacheInvalidationStrategy):
    async def process_update(self, **kwargs: Any) -> None:
        pass

    async def process_retrieve(self, **kwargs: Any) -> None:
        pass

    @property
    def is_cache_disabled(self) -> bool:
        return True


class CacheInvalidationByTimerStrategy(CacheInvalidationStrategy):
    def __init__(self, cache_time_in_seconds: Union[float, int] = INFINITE):
        self._cache_time = cache_time_in_seconds

    @property
    def is_cache_disabled(self) -> bool:
        return self._cache_time == 0

    async def process_update(self, **kwargs: Any) -> None:
        if self.is_cache_disabled:
            raise CacheValidationError()

    async def process_retrieve(self, **kwargs: Any) -> None:
        self._is_cache_expired(**kwargs)

    def _is_cache_expired(self, **kwargs: Dict[str, Union[Any, float]]) -> None:
        for key, value in kwargs.items():
            added_at: float = value[ADD_TIME_PLACEHOLDER]
            elapsed_time_since_adding = time.monotonic() - added_at
            if elapsed_time_since_adding >= self._cache_time:
                raise CacheExpiredError()


class APIResponsesCacheInvalidationStrategy(CacheInvalidationByTimerStrategy):
    _validation_criteria = ('params', 'json', 'data', 'headers')

    def __init__(self, cache_time_in_seconds: Union[float, int] = INFINITE):
        super().__init__(cache_time_in_seconds)
        self._uncached = ('https://api.qiwi.com/partner/bill', '/sinap/api/v2/terms/')

    @property
    def is_cache_disabled(self) -> bool:
        return self._cache_time == 0

    async def process_update(self, **kwargs: Any) -> None:
        await super().process_update(**kwargs)
        for key in kwargs.keys():
            if any(key.startswith(coincidence) for coincidence in self._uncached):
                raise CacheValidationError()

    async def check_is_contains_similar(self, storage: CacheStorage, item: Any) -> bool:
        values = await storage.retrieve_all()
        if not isinstance(item, Payload):
            return item in values
        for value in values:
            if not isinstance(value, CachedAPIRequest):
                continue
            if (
                value.method == 'GET'
                and value.payload.headers == item.headers  # noqa: W503
                and value.payload.params == item.params  # noqa: W503
            ):
                return True
            else:
                if value.method == 'GET':
                    return False
                keys = (k for k in self._validation_criteria if k != 'headers')
                return any(getattr(item, key) == getattr(value.payload, key) for key in keys)
        return False



from typing import ClassVar

from glQiwiApi.core.abc.api_method import APIMethod
from glQiwiApi.yoo_money.types import AccountInfo


class RetrieveAccountInfo(APIMethod[AccountInfo]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://yoomoney.ru/api/account-info'



import uuid
from typing import Any, ClassVar, Dict, Optional

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import PaymentDetails, PaymentInfo, PaymentMethod
from glQiwiApi.types.amount import AmountWithCurrency


class MakePaymentByDetails(QiwiAPIMethod[PaymentInfo]):
    url: ClassVar[str] = 'https://edge.qiwi.com/sinap/api/v2/terms/1717/payments'
    http_method: ClassVar[str] = 'POST'

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        'id': RuntimeValue(default_factory=lambda: str(uuid.uuid4())),
        'sum': RuntimeValue(),
        'paymentMethod': RuntimeValue(),
        'fields': RuntimeValue(),
    }

    payment_sum: AmountWithCurrency = Field(..., scheme_path='sum')
    payment_method: PaymentMethod = Field(..., scheme_path='paymentMethod')
    details: PaymentDetails = Field(..., scheme_path='fields')
    payment_id: Optional[str] = Field(None, scheme_path='id')



import abc
import types
from http import HTTPStatus
from json import JSONDecodeError
from typing import Any, ClassVar, Generic, Sequence, TypeVar, cast

from glQiwiApi.core.abc.api_method import APIMethod, ReturningType, _sentinel
from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.qiwi.exceptions import QiwiAPIError

try:
    from orjson import JSONDecodeError as OrjsonDecodeError
except ImportError:
    from json import JSONDecodeError as OrjsonDecodeError

T = TypeVar('T', bound=Any)


class QiwiAPIMethod(APIMethod[T], abc.ABC, Generic[T]):
    arbitrary_allowed_response_status_codes: ClassVar[Sequence[int]] = ()

    @classmethod
    def parse_http_response(cls, response: HTTPResponse) -> ReturningType:
        response_is_successful = cls.check_if_response_status_success(response)

        try:
            json_response = response.json()
        except (JSONDecodeError, TypeError, OrjsonDecodeError):
            response_is_successful = False

        if not response_is_successful:
            QiwiAPIError(response).raise_exception_matching_error_code()

        # micro optimization that helps to avoid json re-deserialization
        response.json = types.MethodType(lambda self: json_response, response)  # type: ignore

        manually_parsed_json = cls.on_json_parse(response)
        if manually_parsed_json is not _sentinel:
            return cast(ReturningType, manually_parsed_json)

        return super().parse_http_response(response)

    @classmethod
    def check_if_response_status_success(cls, response: HTTPResponse) -> bool:
        if response.status_code == HTTPStatus.OK:
            return True
        elif response.status_code in cls.arbitrary_allowed_response_status_codes:
            return True
        return False



from typing import ClassVar

from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import WebhookInfo


class GetCurrentWebhook(QiwiAPIMethod[WebhookInfo]):
    url: ClassVar[str] = 'https://edge.qiwi.com/payment-notifier/v1/hooks/active'
    http_method: ClassVar[str] = 'GET'



from typing import Any, ClassVar, Dict

from glQiwiApi.qiwi.base import QiwiAPIMethod


class SendTestWebhookNotification(QiwiAPIMethod[Dict[Any, Any]]):
    http_method: ClassVar[str] = 'GET'
    url: ClassVar[str] = 'https://edge.qiwi.com/payment-notifier/v1/hooks/test'



from typing import Any, ClassVar, Dict

from pydantic import Field

from glQiwiApi.core.abc.api_method import APIMethod


class MakeCellularPayment(APIMethod[Dict[str, Any]]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://yoomoney.ru/api/operation-details'

    pattern_id: str
    phone_number: str = Field(..., alias='phone-number')
    amount: float



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, cast

from glQiwiApi.core.event_fetching.class_based.base import Handler
from glQiwiApi.qiwi.clients.p2p.types import Bill
from glQiwiApi.types.amount import PlainAmount

if TYPE_CHECKING:
    from glQiwiApi.qiwi.clients.p2p.client import QiwiP2PClient


class AbstractBillHandler(Handler[Bill], abc.ABC):
    @property
    def wallet(self) -> QiwiP2PClient:
        return cast(QiwiP2PClient, self.context['wallet'])

    @property
    def bill_id(self) -> str:
        return self.event.id

    @property
    def bill_sum(self) -> PlainAmount:
        return self.event.amount

    @property
    def pay_url(self) -> str:
        return self.event.pay_url

    @property
    def shim_url(self) -> str:
        return self.wallet.create_shim_url(self.event.invoice_uid)



from typing import Any, Dict, Optional

from pydantic import Field, Json

from glQiwiApi.types.base import Base


class Code(Base):
    value: str
    name: str = Field(..., alias='_name')


class MobileOperator(Base):
    code: Code
    data: Optional[Json]
    message: str
    messages: Optional[Dict[str, Any]]



from glQiwiApi.core.event_fetching.webhooks.views.base import BaseWebhookView

from . import app  # noqa
from .views import QiwiBillWebhookView, QiwiTransactionWebhookView

__all__ = ('QiwiBillWebhookView', 'QiwiTransactionWebhookView', 'BaseWebhookView', 'app.py')



import time
from typing import Any, ClassVar, Dict, Union

from pydantic import Field

from glQiwiApi.core.abc.api_method import Request, RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import PaymentInfo


class TransferMoneyToCard(QiwiAPIMethod[PaymentInfo]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://edge.qiwi.com/sinap/api/v2/terms/{private_card_id}/payments'

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        'id': RuntimeValue(default_factory=lambda: str(int(time.time() * 1000))),
        'sum': {'amount': RuntimeValue(), 'currency': '643'},
        'paymentMethod': {'type': 'Account', 'accountId': '643'},
        'fields': {'account': RuntimeValue()},
    }

    private_card_id: str = Field(..., path_runtime_value=True)
    amount: float = Field(..., scheme_path='sum.amount')
    card_number: str = Field(..., scheme_path='fields.account')
    kwargs: Dict[str, Any] = {}  # parameters for cards with ID 1960, 21012

    def build_request(self, **url_format_kw: Any) -> 'Request':
        request_schema = self._get_filled_json_payload_schema()
        request_schema['fields'].update(**self.kwargs)
        return Request(
            json_payload=request_schema,
            endpoint=self.url.format(**url_format_kw, **self._get_runtime_path_values()),
            http_method=self.http_method,
        )



from typing import ClassVar

from pydantic import Field

from glQiwiApi.core.abc.api_method import ReturningType
from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.qiwi.base import QiwiAPIMethod


class GenerateWebhookSecret(QiwiAPIMethod[str]):
    url: ClassVar[str] = 'https://edge.qiwi.com/payment-notifier/v1/hooks/{hook_id}/newkey'
    http_method: ClassVar[str] = 'POST'

    hook_id: str = Field(..., path_runtime_value=True)

    @classmethod
    def on_json_parse(cls, response: HTTPResponse) -> ReturningType:
        return response.json()['key']



from typing import List

from pydantic import Field

from glQiwiApi.types.amount import AmountWithCurrency
from glQiwiApi.types.base import Base


class Statistic(Base):
    """object: Statistic"""

    incoming: List[AmountWithCurrency] = Field(alias='incomingTotal')
    out: List[AmountWithCurrency] = Field(alias='outgoingTotal')


__all__ = ['Statistic']



from ipaddress import IPv4Address, IPv4Network
from typing import Optional, Sequence, Set, Union

DEFAULT_QIWI_NETWORKS = [
    IPv4Network('79.142.16.0/20'),
    IPv4Network('195.189.100.0/22'),
    IPv4Network('91.232.230.0/23'),
    IPv4Network('91.213.51.0/24'),
]


class IPFilter:
    def __init__(self, ips: Optional[Sequence[Union[str, IPv4Network, IPv4Address]]] = None):
        self._allowed_ips: Set[IPv4Address] = set()

        if ips:
            self.allow(*ips)

    def allow(self, *ips: Union[str, IPv4Network, IPv4Address]) -> None:
        for ip in ips:
            self.allow_ip(ip)

    def allow_ip(self, ip: Union[str, IPv4Network, IPv4Address]) -> None:
        if isinstance(ip, str):
            ip = IPv4Network(ip) if '/' in ip else IPv4Address(ip)
        if isinstance(ip, IPv4Address):
            self._allowed_ips.add(ip)
        elif isinstance(ip, IPv4Network):
            self._allowed_ips.update(ip.hosts())
        else:
            raise ValueError(f"Invalid type of ipaddress: {type(ip)} ('{ip}')")

    @classmethod
    def default(cls) -> 'IPFilter':
        return cls(DEFAULT_QIWI_NETWORKS)

    def __contains__(self, item: Union[str, IPv4Address]) -> bool:
        return self.check(item)

    def check(self, ip: Union[str, IPv4Address]) -> bool:
        if not isinstance(ip, IPv4Address):
            ip = IPv4Address(ip)
        return ip in self._allowed_ips



from datetime import datetime, timezone
from typing import Optional

DEFAULT_QIWI_TIMEZONE = 'Europe/Moscow'

try:
    import zoneinfo
except ImportError:
    import backports.zoneinfo as zoneinfo


def datetime_to_utc_in_iso_format(obj: datetime) -> str:
    return obj.astimezone(tz=timezone.utc).isoformat(timespec='milliseconds')


def datetime_to_iso8601_with_moscow_timezone(obj: Optional[datetime]) -> str:
    """
    Converts a date to a standard format for API's

    :param obj: datetime object to parse to string
    :return: string - parsed date
    """
    if not isinstance(obj, datetime):
        return ''  # pragma: no cover
    return obj.astimezone(zoneinfo.ZoneInfo(DEFAULT_QIWI_TIMEZONE)).isoformat(timespec='seconds')



import abc
from typing import Any, Dict


class Pluggable(abc.ABC):
    """Something pluggable into library"""

    @abc.abstractmethod
    async def install(self, ctx: Dict[Any, Any]) -> None:
        ...

    @abc.abstractmethod
    async def shutdown(self) -> None:
        ...



import uuid
from typing import ClassVar, Optional

from pydantic import Field

from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types.qiwi_master import QiwiMasterCardRequisites


class GetQiwiMasterCardRequisites(QiwiAPIMethod[QiwiMasterCardRequisites]):
    http_method: ClassVar[str] = 'PUT'
    url: ClassVar[str] = 'https://edge.qiwi.com/cards/v1/cards/{card_id}/details'

    card_id: str = Field(..., path_runtime_value=True)
    operation_id: Optional[str] = Field(
        default_factory=lambda: str(uuid.uuid4()), alias='operationId'
    )



from typing import ClassVar

from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import Identification


class GetIdentification(QiwiAPIMethod[Identification]):
    http_method: ClassVar[str] = 'GET'
    url: ClassVar[
        str
    ] = 'https://edge.qiwi.com/identification/v1/persons/{phone_number}/identification'



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Dict, Generic, TypeVar, Union

if TYPE_CHECKING:
    from glQiwiApi.types.base import Base  # NOQA # pragma: no cover

T = TypeVar('T', bound=Union[Exception, 'Base'])


class BaseHandlerMixin(Generic[T]):
    if TYPE_CHECKING:  # pragma: no cover
        event: T


class Handler(abc.ABC, BaseHandlerMixin[T]):
    """Base class for all class-based handlers"""

    def __init__(self, event: T, *args: Any) -> None:
        self.event: T = event
        self._args = args
        if args:
            self.context: Dict[str, Any] = args[0].context

    @abc.abstractmethod
    async def process_event(self) -> Any:  # pragma: no cover  # type: ignore
        raise NotImplementedError

    def __await__(self) -> Any:
        return self.process_event().__await__()



from typing import Any, ClassVar, Dict, Optional

from pydantic import Field

from glQiwiApi.qiwi.base import QiwiAPIMethod


class AuthenticateWallet(QiwiAPIMethod[Dict[str, Any]]):
    url: ClassVar[
        str
    ] = 'https://edge.qiwi.com/identification/v1/persons/{phone_number}/identification'
    http_method: ClassVar[str] = 'POST'

    passport: str
    birth_date: str = Field(..., alias='birthDate')
    first_name: str = Field(..., alias='firstName')
    middle_name: str = Field(..., alias='middleName')
    last_name: str = Field(..., alias='lastName')
    inn: str = ''
    snils: str = ''
    oms: str = ''



from typing import Any, ClassVar, Dict, List, Optional

from pydantic import Field

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.maps.types.polygon import Polygon
from glQiwiApi.qiwi.clients.maps.types.terminal import Terminal


class GetTerminals(QiwiAPIMethod[List[Terminal]]):
    url: ClassVar[str] = 'http://edge.qiwi.com/locator/v3/nearest/clusters?parameters'
    http_method: ClassVar[str] = 'GET'

    polygon: Polygon
    pop_if_inactive_x_mins: int = Field(..., alias='activeWithinMinutes')
    zoom: Optional[int] = None
    include_partners: Optional[bool] = Field(None, alias='withRefillWallet')
    partners_ids: Optional[List[Any]] = Field(None, alias='ttpIds')
    cache_terminals: Optional[bool] = Field(None, alias='cacheAllowed')
    card_terminals: Optional[bool] = Field(None, alias='cardAllowed')
    identification_types: Optional[int] = Field(None, alias='identificationTypes')
    terminal_groups: Optional[List[Any]] = Field(
        None,
        alias='ttpGroups',
    )

    def build_request(self, **url_format_kw: Any) -> 'Request':
        model_dict = self.dict(exclude_none=True, exclude_unset=True, by_alias=True)
        polygon = model_dict.pop('polygon')

        model_dict = _replace_bool_values_with_strings(model_dict)

        return Request(
            endpoint=self.url.format(**url_format_kw, **self._get_runtime_path_values()),
            http_method=self.http_method,
            params={**model_dict, **polygon},
        )


def _replace_bool_values_with_strings(d: Dict[str, Any]) -> Dict[str, Any]:
    for k, v in d.items():
        if not isinstance(v, bool):
            continue

        d[k] = str(v)
    return d



import ipaddress
import pathlib
import ssl
from dataclasses import dataclass
from datetime import datetime, timedelta
from io import BytesIO
from typing import TYPE_CHECKING, Any, Iterable, List, Optional, Union, cast

if TYPE_CHECKING:
    try:
        from cryptography import x509  # NOQA  # pragma: no cover
        from cryptography.hazmat.primitives.asymmetric.rsa import (
            RSAPrivateKeyWithSerialization,  # pragma: no cover; NOQA; NOQA
        )
    except ImportError:  # pragma: no cover
        pass

try:
    from aiogram.types import InputFile  # NOQA  # type: ignore
except ImportError:  # pragma: no cover
    pass


def get_or_generate_self_signed_certificate(
    hostname: str,  # your host machine ip address
    cert_path: Union[str, pathlib.Path] = 'cert.pem',
    pkey_path: Union[str, pathlib.Path] = 'pkey.pem',
    ip_addresses: Optional[Iterable[Any]] = None,
    rsa_private_key: Optional['RSAPrivateKeyWithSerialization'] = None,
    public_exponent: int = 65537,
    key_size: int = 2048,
    backend: Optional[Any] = None,
    serial_number: int = 1000,
    expire_days: int = 3650,
    *name_attributes: 'x509.NameAttribute',
) -> 'SSLCertificate':
    """
    Generates self signed certificate for a hostname, and optional IP addresses.

    Copied from https://gist.github.com/bloodearnest/9017111a313777b9cce5
    """
    try:
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import hashes, serialization
        from cryptography.hazmat.primitives.asymmetric import rsa
        from cryptography.x509.oid import NameOID
    except ImportError:  # pragma: no cover
        raise ImportError(  # pragma: no cover
            'You need to install cryptography package for generating self-signed cert'
        )

    if pathlib.Path(cert_path).is_file() and pathlib.Path(pkey_path).is_file():
        return SSLCertificate(_cert_path=cert_path, _pkey_path=pkey_path)  # pragma: no cover

    # Generate our key
    if rsa_private_key is None:
        rsa_private_key = rsa.generate_private_key(
            public_exponent=public_exponent,
            key_size=key_size,
            backend=backend or default_backend(),
        )

    name = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, hostname), *name_attributes])

    # best practice seem to be to include the hostname in the SAN,
    # which *SHOULD* mean COMMON_NAME is ignored.
    alt_names: List[x509.GeneralName] = [x509.DNSName(hostname)]

    # allow addressing by IP, for when you don't have real DNS (common in most testing scenarios
    if ip_addresses:
        for address in ip_addresses:
            # openssl wants DNSnames for ips...
            alt_names.append(x509.DNSName(address))
            # ... whereas golang's crypto/tls is stricter, and needs IPAddresses
            # note: older versions of cryptography do not understand ip_address objects
            alt_names.append(x509.IPAddress(ipaddress.ip_address(address)))

    san = x509.SubjectAlternativeName(alt_names)

    # path_len=0 means this cert can only sign itself, not other certs.
    basic_constraints = x509.BasicConstraints(ca=True, path_length=0)
    now = datetime.utcnow()
    cert = (
        x509.CertificateBuilder()
        .subject_name(name)
        .issuer_name(name)
        .public_key(rsa_private_key.public_key())
        .serial_number(serial_number)
        .not_valid_before(now)
        .not_valid_after(now + timedelta(days=expire_days))
        .add_extension(basic_constraints, False)
        .add_extension(san, False)
        .sign(rsa_private_key, hashes.SHA256(), default_backend())
    )
    cert_pem = cert.public_bytes(encoding=cast(serialization.Encoding, serialization.Encoding.PEM))
    key_pem = rsa_private_key.private_bytes(
        encoding=cast(serialization.Encoding, serialization.Encoding.PEM),
        format=cast(
            serialization.PrivateFormat,
            serialization.PrivateFormat.TraditionalOpenSSL,
        ),
        encryption_algorithm=serialization.NoEncryption(),
    )

    with open(cert_path, 'wb+') as f1, open(pkey_path, 'wb') as f2:
        f1.write(cert_pem)
        f2.write(key_pem)
    return SSLCertificate(_cert_path=cert_path, _pkey_path=pkey_path)


@dataclass
class SSLCertificate:
    _cert_path: Union[str, pathlib.Path]
    _pkey_path: Union[str, pathlib.Path]

    def as_ssl_context(self) -> ssl.SSLContext:
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        context.load_cert_chain(self._cert_path, self._pkey_path)
        return context

    def as_input_file(self) -> 'InputFile':
        with open(self._cert_path, 'rb') as file:
            return InputFile(BytesIO(file.read()))



import uuid
from datetime import datetime, timedelta
from typing import Any, ClassVar, Dict, List, Optional, Union

from pydantic import Field, validator

from glQiwiApi.core.abc.api_method import Request, RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.p2p.types import Bill, Customer
from glQiwiApi.utils.date_conversion import datetime_to_iso8601_with_moscow_timezone


def get_default_bill_life_time() -> datetime:
    return datetime.now() + timedelta(days=2)


class CreateP2PBill(QiwiAPIMethod[Bill]):
    http_method: ClassVar[str] = 'PUT'
    url: ClassVar[str] = 'https://api.qiwi.com/partner/bill/v1/bills/{bill_id}'

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        'amount': {'currency': 'RUB', 'value': RuntimeValue()},
        'expirationDateTime': RuntimeValue(default_factory=get_default_bill_life_time),
        'comment': RuntimeValue(mandatory=False),
        'customFields': {
            'paySourcesFilter': RuntimeValue(default=['qw', 'card']),
            'themeCode': RuntimeValue(default='Yvan-YKaSh'),
        },
        'customer': RuntimeValue(mandatory=False),
    }

    bill_id: Optional[str] = Field(None, path_runtime_value=True)
    expire_at: Optional[datetime] = Field(
        scheme_path='expirationDateTime',
    )
    amount: float = Field(..., scheme_path='amount.value')
    comment: Optional[str] = Field(None, scheme_path='comment')
    theme_code: Optional[str] = Field(None, scheme_path='customFields.themeCode')
    pay_source_filter: Optional[List[str]] = Field(
        None, scheme_path='customFields.paySourcesFilter'
    )
    customer: Optional[Customer] = Field(None, schema_path='customer')

    @validator('bill_id')
    def set_bill_id_if_it_is_none(cls, v: Optional[str]) -> str:
        return v or str(uuid.uuid4())

    def build_request(self, **url_format_kw: Any) -> 'Request':
        request = super().build_request(**url_format_kw)
        expire_at = request.json_payload['expirationDateTime']  # type: ignore
        request.json_payload['expirationDateTime'] = datetime_to_iso8601_with_moscow_timezone(
            expire_at
        )
        pay_source_filter = request.json_payload['customFields']['paySourcesFilter']
        if isinstance(pay_source_filter, list):
            request.json_payload['customFields']['paySourcesFilter'] = ' '.join(
                pay_source_filter
            ).replace(' ', ',')
        return request



from typing import Any, ClassVar, Dict

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod


class RenameQiwiMasterCard(QiwiAPIMethod[Dict[str, Any]]):
    url: ClassVar[str] = 'https://edge.qiwi.com/cards/v1/cards/{card_id}>/alias'
    http_method: ClassVar[str] = 'PUT'

    json_payload_schema = {'alias': RuntimeValue()}

    card_id: str = Field(..., path_runtime_value=True)
    alias: str



from typing import ClassVar, List

from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types.restriction import Restriction


class GetRestrictions(QiwiAPIMethod[List[Restriction]]):
    url: ClassVar[
        str
    ] = 'https://edge.qiwi.com/person-profile/v1/persons/{phone_number}/status/restrictions'
    http_method: ClassVar[str] = 'GET'



from __future__ import annotations

import abc
from dataclasses import dataclass
from types import TracebackType
from typing import Any, Dict, Generic, Mapping, Optional, Set, Type, TypeVar, cast

import aiohttp
from aiohttp import ClientResponse

from glQiwiApi.utils.compat import json

_SessionType = TypeVar('_SessionType', bound=Any)
_SessionHolderType = TypeVar('_SessionHolderType', bound='AbstractSessionHolder[Any]')


@dataclass
class HTTPResponse:
    status_code: int
    body: bytes
    headers: Mapping[str, Any]
    content_type: str

    def json(self) -> Dict[str, Any]:
        return cast(Dict[str, Any], json.loads(self.body))


class AbstractSessionHolder(abc.ABC, Generic[_SessionType]):
    """
    Manages the lifecycle of the session (s) and allows spoofing
    library for requests, for example from aiohttp to httpx without any problem.
    Holder is lazy and allocates in his own state
    session on first call, not on instantiation.
    """

    def __init__(self, **kwargs: Any) -> None:
        self._session: Optional[_SessionType] = None
        self._session_kwargs = kwargs

    @abc.abstractmethod
    async def convert_third_party_lib_response_to_http_response(
        self, response: Any
    ) -> HTTPResponse:
        raise NotImplementedError

    def update_session_kwargs(self, **kwargs: Any) -> None:
        self._session_kwargs.update(kwargs)

    async def __aenter__(self: AbstractSessionHolder[_SessionType]) -> _SessionType:
        self._session = await self.get_session()
        return self._session

    @abc.abstractmethod
    async def get_session(self) -> _SessionType:
        raise NotImplementedError

    async def __aexit__(
        self: AbstractSessionHolder[_SessionType],
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()

    @abc.abstractmethod
    async def close(self) -> None:
        raise NotImplementedError


class AiohttpSessionHolder(AbstractSessionHolder[aiohttp.ClientSession]):
    def __init__(self, **kwargs: Any):
        AbstractSessionHolder.__init__(self, **kwargs)

    async def close(self) -> None:
        if self._session_in_working_order():
            await self._session.close()

    async def convert_third_party_lib_response_to_http_response(
        self, response: ClientResponse
    ) -> HTTPResponse:
        return HTTPResponse(
            status_code=response.status,
            body=await response.read(),
            headers=response.headers,
            content_type=response.content_type,
        )

    async def get_session(self) -> _SessionType:
        if self._session_in_working_order():
            return self._session
        return await self._instantiate_new_session()

    def _session_in_working_order(self) -> bool:
        return self._session is not None and self._session.closed is False

    async def _instantiate_new_session(self) -> _SessionType:
        self._session: _SessionType = cast(
            _SessionType, aiohttp.ClientSession(**self._session_kwargs)
        )
        return self._session



from glQiwiApi.utils.synchronous import async_as_sync, execute_async_as_sync

from .abc.base_api_client import BaseAPIClient
from .event_fetching.class_based import (
    AbstractBillHandler,
    AbstractTransactionHandler,
    AbstractTransactionWebhookHandler,
    ErrorHandler,
    Handler,
)
from .event_fetching.filters import BaseFilter, LambdaBasedFilter
from .event_fetching.webhooks import (
    BaseWebhookView,
    QiwiBillWebhookView,
    QiwiTransactionWebhookView,
    app,
)
from .request_service import RequestService

__all__ = (
    'RequestService',
    'BaseFilter',
    'LambdaBasedFilter',
    # class-based handlers
    'Handler',
    'AbstractBillHandler',
    'AbstractTransactionHandler',
    'AbstractTransactionWebhookHandler',
    'ErrorHandler',
    # synchronous adapters and utils
    'async_as_sync',
    'execute_async_as_sync',
    # webhooks
    'QiwiBillWebhookView',
    'QiwiTransactionWebhookView',
    'BaseWebhookView',
)



import http
from typing import Any, ClassVar, Dict, Sequence

from pydantic import Field

from glQiwiApi.core.abc.api_method import ReturningType
from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.qiwi.base import QiwiAPIMethod


class BlockQiwiMasterCard(QiwiAPIMethod[Dict[str, Any]]):
    url: ClassVar[str] = '/cards/v2/persons/{phone_number}/cards/{card_id}/block'
    http_method: ClassVar[str] = 'GET'

    arbitrary_allowed_response_status_codes: ClassVar[Sequence[int]] = (http.HTTPStatus.ACCEPTED,)

    card_id: str = Field(..., path_runtime_value=True)
    phone_number: str = Field(..., path_runtime_value=True)

    @classmethod
    def parse_http_response(cls, response: HTTPResponse) -> Dict[str, Any]:
        return {}



from typing import Any, ClassVar, Dict, List, Sequence

from glQiwiApi.core.abc.api_method import Request, ReturningType
from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import Limit

ALL_LIMIT_TYPES = [
    'TURNOVER',
    'REFILL',
    'PAYMENTS_P2P',
    'PAYMENTS_PROVIDER_INTERNATIONALS',
    'PAYMENTS_PROVIDER_PAYOUT',
    'WITHDRAW_CASH',
]


class GetLimits(QiwiAPIMethod[Dict[str, Limit]]):
    http_method: ClassVar[str] = 'GET'
    url: ClassVar[str] = 'https://edge.qiwi.com/qw-limits/v1/persons/{phone_number}/actual-limits'

    limit_types: Sequence[str] = ALL_LIMIT_TYPES

    @classmethod
    def parse_http_response(cls, response: HTTPResponse) -> Dict[str, Limit]:
        return {
            code: [Limit.parse_obj(limit) for limit in limits]
            for code, limits in response.json()['limits'].items()
        }

    def build_request(self, **url_format_kw: Any) -> Request:
        return Request(
            endpoint=self.url.format(**url_format_kw),
            params={
                f'types[{index}]': limit_type for index, limit_type in enumerate(self.limit_types)
            },
            http_method=self.http_method,
        )



"""
Provides effortless work with YooMoney API using asynchronous requests.

"""
from __future__ import annotations

import asyncio
import typing
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Union

from glQiwiApi.core.abc.base_api_client import BaseAPIClient, RequestServiceFactoryType
from glQiwiApi.core.request_service import RequestService, RequestServiceProto
from glQiwiApi.core.session import AiohttpSessionHolder
from glQiwiApi.utils.payload import make_payload
from glQiwiApi.utils.validators import String
from glQiwiApi.yoo_money.methods.acccept_incoming_transfer import AcceptIncomingTransfer
from glQiwiApi.yoo_money.methods.build_auth_url import BuildAuthURL
from glQiwiApi.yoo_money.methods.get_access_token import GetAccessToken
from glQiwiApi.yoo_money.methods.make_cellular_payment import MakeCellularPayment
from glQiwiApi.yoo_money.methods.operation_details import OperationDetailsMethod
from glQiwiApi.yoo_money.methods.operation_history import OperationHistoryMethod
from glQiwiApi.yoo_money.methods.process_payment import ProcessPayment
from glQiwiApi.yoo_money.methods.reject_incoming_transfer import RejectIncomingTransfer
from glQiwiApi.yoo_money.methods.request_payment import RequestPayment
from glQiwiApi.yoo_money.methods.retrieve_account_info import RetrieveAccountInfo
from glQiwiApi.yoo_money.methods.revoke_api_token import RevokeAPIToken
from glQiwiApi.yoo_money.types import (
    AccountInfo,
    IncomingTransaction,
    OperationDetails,
    OperationHistory,
    Payment,
    RequestPaymentResponse,
)


class YooMoneyAPI(BaseAPIClient):
    """
    That class implements processing requests to YooMoney
    It is convenient in that it does not just give json such objects,
    and all this converts into pydantic models.
    To work with this class, you need to register a token,
    using the guide on the official github of the project

    """

    _api_access_token = String(optional=False)

    def __init__(
        self,
        api_access_token: str,
        request_service_factory: Optional[RequestServiceFactoryType] = None,
    ) -> None:
        """
        The constructor accepts a token obtained from the method class get_access_token
         and the special attribute without_context

        :param api_access_token: api token for requests
        """
        self._api_access_token = api_access_token
        BaseAPIClient.__init__(self, request_service_factory)

    async def _create_request_service(self) -> RequestServiceProto:
        return RequestService(
            session_holder=AiohttpSessionHolder(
                headers={
                    'Content-Type': 'application/x-www-form-urlencoded',
                    'Authorization': f'Bearer {self._api_access_token}',
                    'Host': 'yoomoney.ru',
                }
            )
        )

    @classmethod
    async def build_url_for_auth(
        cls, scopes: List[str], client_id: str, redirect_uri: str = 'https://example.com'
    ) -> str:
        """
        Method to get the link for further authorization and obtaining a token

        :param scopes: OAuth2 authorization of the application by the user,
         the rights are transferred by the list.
        :param client_id: application id, type string
        :param redirect_uri: a funnel where the temporary code that you need will go to
         to get the main token
        :return: the link to follow
         and make authorization via login / password
        """
        request_service = RequestService()
        request = BuildAuthURL(
            client_id=client_id, scopes=scopes, redirect_uri=redirect_uri
        ).build_request()
        try:
            return BuildAuthURL.parse_http_response(
                await request_service.send_request(
                    request.endpoint,
                    request.http_method,
                    params=request.params,
                    data=request.data,
                    headers=request.headers,
                )
            )
        finally:
            await request_service.shutdown()

    @classmethod
    async def get_access_token(
        cls,
        code: str,
        client_id: str,
        redirect_uri: str = 'https://example.com',
        client_secret: Optional[str] = None,
    ) -> str:
        """
        Method for obtaining a token for requests to the YooMoney API

        :param code: the temporary code that was obtained in the base_authorize method
        :param client_id: application id, type string
        :param redirect_uri: the funnel where the temporary code will go,
         which is needed to get the main token
        :param client_secret: The secret word for authenticating the application.
         Specified if the service is registered with authentication.
        :return: YooMoney API TOKEN
        """
        rs = RequestService()
        try:
            return await rs.execute_api_method(
                GetAccessToken(
                    code=code,
                    client_id=client_id,
                    redirect_uri=redirect_uri,
                    client_secret=client_secret,
                )
            )
        finally:
            await rs.shutdown()

    async def revoke_api_token(self) -> None:
        """
        Method for revoking the rights of a token, while all its rights also disappear
        Documentation:
        https://yoomoney.ru/docs/wallet/using-api/authorization/revoke-access-token
        """
        await self._request_service.execute_api_method(RevokeAPIToken())

    @classmethod
    def create_pay_form(
        cls,
        receiver: str,
        quick_pay_form: str,
        targets: str,
        payment_type: str,
        amount: Union[int, float],
        form_comment: Optional[str] = None,
        short_dest: Optional[str] = None,
        label: Optional[str] = None,
        comment: Optional[str] = None,
        success_url: Optional[str] = None,
        need_fio: Optional[bool] = None,
        need_email: Optional[bool] = None,
        need_phone: Optional[bool] = None,
        need_address: Optional[bool] = None,
    ) -> str:
        """
        The YooMoney form is a set of fields with information about a transfer.
        You can embed payment form into your interface (for instance, a website or blog).
        When the sender pushes the button, the details from the form are sent to YooMoney
        and an order for a transfer to your wallet is initiated.

        Detail docs: https://yoomoney.ru/docs/payment-buttons/using-api/forms?lang=en

        Possible values for quick_pay_form:
           * shop - for a multi purpose form;
           * small - for a button;
           * donate - for a charity form.

        @param receiver: Number of the YooMoney wallet which money from senders is credited to.
        @param quick_pay_form:
        @param targets: Payment purpose.
        @param payment_type: Payment method. Possible values: PC, AC, MC
        @param amount: 	Transfer amount (the amount debited from the sender).
        @param form_comment: Name of the transfer in sender’s history
                (for transfers from a wallet or linked bank card). Displayed in sender’s wallet.
                The simplest way to create it is to combine the names of the store and product.
                For instance:  My Store: white valenki boots
        @param short_dest: Name of the transfer on the confirmation page.
               We recommend using the same name as formcomment
        @param label: The label that a site or app assigns to a certain transfer.
                      For instance, a code or order identifier may be used for this label.
        @param comment: The field in which you can transfer_money sender’s comments.
        @param success_url: URL where the user is redirected after the transfer.
        @param need_fio: Sender’s full name required.
        @param need_email: Sender’s email address required.
        @param need_phone: Sender’s phone number required.
        @param need_address: Sender’s address required.
        @return: link to payment form
        """
        payload = make_payload(
            **{
                'receiver': receiver,
                'quickpay-form': quick_pay_form,
                'targets': targets,
                'paymentType': payment_type,
                'sum': amount,
                'formcomment': form_comment,
                'short-dest': short_dest,
                'label': label,
                'comment': comment,
                'successURL': success_url,
                'need-fio': need_fio,
                'need-email': need_email,
                'need-phone': need_phone,
                'need-address': need_address,
            }
        )
        base_url = 'https://yoomoney.ru/quickpay/confirm.xml?'
        params = ''.join(f'&{key}={value}' for key, value in payload.items())
        return base_url + params

    async def retrieve_account_info(self) -> AccountInfo:
        """
        Method for getting information about user account
        Detailed documentation:
        https://yoomoney.ru/docs/wallet/user-account/account-info

        :return: объект AccountInfo
        """
        return await self._request_service.execute_api_method(RetrieveAccountInfo())

    async def operation_history(
        self,
        operation_types: Optional[Iterable[str]] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        start_record: Optional[int] = None,
        records: int = 30,
        label: Optional[Union[str, int]] = None,
        in_detail: bool = False,
    ) -> OperationHistory:
        """
        More details:
        https://yoomoney.ru/docs/wallet/user-account/operation-history\n
        The method allows you to view the history of operation_history (in whole or in part) in page mode.
        History records are displayed in reverse chronological order from most recent to earlier.

        Possible values: \n
        DEPOSITION — refill ;\n
        PAYMENT — consumption;\n
        INCOMING(incoming-transfers-unaccepted) — unaccepted incoming P2P transfers of any type.\n

        :param operation_types: Operation type
        :param label: string.
         Selection of payments by tag value.
         Selects payments that have a specified parameter value
         label of the request-payment call.
        :param start_date: Show operations from the moment in time
         (operations equal to start_date or later)
         If the parameter is absent, all operations are displayed.
        :param end_date: Output operations up to the point in time
            (operations older than end_date).
            If the parameter is absent, all operations are displayed.
        :param start_record: If the parameter is present, then operations will be displayed starting
          from start_record number.
         Operations are numbered from 0. More about paginated list output
        :param records:	The number of transaction history records requested.
         Valid values are from 1 to 100, the default is 30.
        :param in_detail:
        """
        return await self._request_service.execute_api_method(
            OperationHistoryMethod(
                operation_types=operation_types,
                start_record=start_record,
                start_date=start_date,
                records=records,
                label=label,
                in_detail=in_detail,
                end_date=end_date,
            )
        )

    async def operation_details(self, operation_id: Union[int, str]) -> OperationDetails:
        """
        Allows you to get detailed information about the operation from the history.
        Required token rights: operation-details.
        More detailed documentation:
        https://yoomoney.ru/docs/wallet/user-account/operation-details

        :param operation_id: Operation ID
        """
        return await self._request_service.execute_api_method(
            OperationDetailsMethod(operation_id=operation_id)
        )

    async def make_cellular_payment(
        self, pattern_id: str, phone_number: str, amount: Union[int, float, str]
    ) -> Dict[str, Any]:
        return await self._request_service.execute_api_method(
            MakeCellularPayment(pattern_id=pattern_id, phone_number=phone_number, amount=amount)
        )

    async def transfer_money(
        self,
        to_account: str,
        amount: Union[int, float],
        pattern_id: str = 'p2p',
        protect: bool = False,
        comment_for_history: Optional[str] = None,
        comment: Optional[str] = None,
        expire_period: int = 1,
    ) -> Payment:
        """
        A method for sending money to another person's account or card.
        This function makes 2 requests at once, because of this you may feel a slight loss in performance,
          you can use the method
        make_request_payment and get the RequestPaymentResponse object,
        which contains information about a still unconfirmed payment \n
        More detailed documentation:
        https://yoomoney.ru/docs/wallet/process-payments/process-payment

        :param pattern_id: Payment pattern ID
        :param to_account: string ID of the transfer recipient
         (account number, phone number or email).
        :param amount: Amount to be received
         (the invoice will be sent to the recipient's account after payment). MINIMUM AMOUNT 2.
         then you will need to pass card_type
         to search for a card in the list of your bank cards,
         and also optionally cvv2 code for making a payment
        :param comment_for_history: Comment to the translation,
         displayed in the sender's history.
        :param comment: Comment on the transfer, displayed to the recipient.
        :param protect: The value of the parameter is true - a sign that
         that the transfer is protected by a protection code.
         By default, there is no parameter (normal translation).
        :param expire_period: Number of days during which:
            the recipient of the transfer can enter the protection code and
            receive a transfer to your account,
            the recipient of the transfer on demand can receive the transfer.
            The parameter value must be in the range from 1 to 365.
            Optional parameter. The default is 1.


        """
        pre_payment = await self.make_request_payment(
            to_account=to_account,
            amount=amount,
            comment_for_history=comment_for_history,
            comment_for_receiver=comment,
            expire_period=expire_period,
            protect=protect,
            pattern_id=pattern_id,
        )
        return await self._request_service.execute_api_method(
            ProcessPayment(request_id=pre_payment.request_id)
        )

    async def make_request_payment(
        self,
        to_account: str,
        amount: Union[int, float],
        pattern_id: str = 'p2p',
        comment_for_history: Optional[str] = None,
        comment_for_receiver: Optional[str] = None,
        protect: bool = False,
        expire_period: int = 1,
    ) -> RequestPaymentResponse:
        """
        More detailed documentation:
        https://yoomoney.ru/docs/wallet/process-payments/request-payment
        Creation of payment, verification of parameters and acceptance
        payment by the store or transfer of funds to the user account of YooMoney.
        This method is not recommended to be used directly, much
        it's easier to use transfer_money.
        Required token rights: to-account ("recipient id", "id type")

        :param pattern_id: Payment pattern ID
        :param to_account: string ID of the transfer recipient
         (account number, phone number or email).
        :param amount: Amount to be received
         (the invoice will be sent to the recipient's account after payment).
        :param comment_for_history: Comment to the translation,
         displayed in the sender's history.
        :param comment_for_receiver: string Comment to the translation,
         displayed to the recipient.
        :param protect: The value of the parameter is true - a sign that
         that the transfer is protected by a protection code.
         By default, there is no parameter (normal translation).
        :param expire_period: The number of days during which:
            the recipient of the transfer can enter the protection code
            and receive a transfer to your account,
            the recipient of the transfer on demand can receive the transfer.
            The parameter value must be in the range from 1 to 365.
            Optional parameter. The default is 1.
        """
        return await self._request_service.execute_api_method(
            RequestPayment(
                to_account=to_account,
                amount=amount,
                pattern_id=pattern_id,
                comment_for_history=comment_for_history,
                comment_for_receiver=comment_for_receiver,
                protect=protect,
                expire_period=expire_period,
            )
        )

    async def get_balance(self) -> float:
        return (await self.retrieve_account_info()).balance

    async def accept_incoming_transaction(
        self, operation_id: str, protection_code: str
    ) -> IncomingTransaction:
        """
        Acceptance of incoming transfers protected by a protection code,
        if you passed the protect parameter to the transfer_money method
        Number of reception attempts
        incoming transfer with a protection code is limited.
        When the number of attempts is exhausted, the transfer is automatically rejected
        (the transfer is returned to the sender).
        More detailed documentation:
        https://yoomoney.ru/docs/wallet/process-payments/incoming-transfer-accept

        :param operation_id: Operation identifier,
         the value of the operation_id parameter of the history () method response
        :param protection_code: Protection code. A string of 4 decimal digits.
         Indicated for an incoming transfer protected by a protection code.
         Not available for on-demand transfers.
        """
        return await self._request_service.execute_api_method(
            AcceptIncomingTransfer(operation_id=operation_id, protection_code=protection_code)
        )

    async def reject_incoming_transfer(self, operation_id: str) -> Dict[str, str]:
        """
        Cancellation of incoming transfers protected by a protection code if you transferred
         in the transfer_money method the protect parameter,
        and transfers on demand. \n
        If the transfer is canceled, it is returned to the sender. \n
        Required token rights: incoming-transfers
        Docs:
        https://yoomoney.ru/docs/wallet/process-payments/incoming-transfer-reject

        :param operation_id: Operation identifier, parameter value
         operation_id of history () method response.
        """
        return await self._request_service.execute_api_method(
            RejectIncomingTransfer(operation_id=operation_id)
        )

    async def check_if_operation_exists(
        self,
        check_fn: typing.Callable[[OperationDetails], bool],
        operation_types: Optional[Iterable[str]] = None,
        max_records: int = 100,
        check_timeout: Optional[float] = None,
    ) -> bool:
        """
        Method for verifying a transaction.
        This method uses self.operation_history (rows = rows) to receive payments.
        For a little optimization, you can decrease rows by setting it,
        however, this does not guarantee the correct result

        :param check_fn:
        :param operation_types: payment type
        :param max_records: number of payments to be checked
        :param check_timeout:
        :return: bool, is there such a transaction in the payment history
        """
        operation_history = await self.operation_history(
            operation_types=operation_types, records=max_records
        )
        get_operation_details_coroutines = [
            self.operation_details(op.id) for op in operation_history
        ]

        for future in asyncio.as_completed(
            get_operation_details_coroutines, timeout=check_timeout
        ):
            if check_fn(await future):
                return True

        return False



import ipaddress
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import Field, validator

from glQiwiApi.types.amount import AmountWithCurrency, CurrencyModel
from glQiwiApi.types.base import Base
from glQiwiApi.utils.currency_util import Currency


class PassInfo(Base):
    """object: PassInfo"""

    last_pass_change: str = Field(alias='lastPassChange')
    next_pass_change: str = Field(alias='nextPassChange')
    password_used: bool = Field(alias='passwordUsed')


class MobilePinInfo(Base):
    """object: MobilePinInfo"""

    last_mobile_pin_change: Optional[datetime] = Field(alias='lastMobilePinChange')
    mobile_pin_used: bool = Field(alias='mobilePinUsed')
    next_mobile_pin_change: str = Field(alias='nextMobilePinChange')


class PinInfo(Base):
    """object: PinInfo"""

    pin_used: bool = Field(alias='pinUsed')


class AuthInfo(Base):
    """object: AuthInfo"""

    ip: ipaddress.IPv4Address
    bound_email: Optional[str] = Field(None, alias='boundEmail')
    last_login_date: Optional[datetime] = Field(None, alias='lastLoginDate')
    email_settings: Optional[Dict[Any, Any]] = Field(None, alias='emailSettings')
    mobile_pin_info: MobilePinInfo = Field(alias='mobilePinInfo')
    pass_info: PassInfo = Field(alias='passInfo')
    person_id: int = Field(alias='personId')
    pin_info: PinInfo = Field(alias='pinInfo')
    registration_date: datetime = Field(alias='registrationDate')


class SmsNotification(Base):
    """object: SmsNotification"""

    price: AmountWithCurrency
    enabled: bool
    active: bool
    end_date: Optional[datetime] = Field(None, alias='endDate')


class IdentificationInfo(Base):
    """object: IdentificationInfo"""

    bank_alias: Optional[str] = Field(None, alias='bankAlias')
    identification_level: Optional[str] = Field(None, alias='identificationLevel')
    passport_expired: Optional[bool] = Field(None, alias='passportExpired')


class NickName(Base):
    """object: NickName"""

    nickname: Optional[str] = None
    can_change: bool = Field(alias='canChange')
    can_use: bool = Field(alias='canUse')
    description: str = ''


class Feature(Base):
    """object: Feature"""

    feature_id: int = Field(alias='featureId')
    feature_value: str = Field(alias='featureValue')
    start_date: str = Field(alias='startDate')
    end_date: str = Field(alias='endDate')


class ContractInfo(Base):
    """object: ContractInfo"""

    blocked: bool = False
    contract_id: int = Field(alias='contractId')
    creation_date: datetime = Field(alias='creationDate')
    identification_info: List[IdentificationInfo] = Field(alias='identificationInfo')
    sms_notification: SmsNotification = Field(alias='smsNotification')
    nickname: NickName
    features: Optional[List[Feature]] = None


class UserInfo(Base):
    """object: UserInfo"""

    default_pay_currency: CurrencyModel = Field(..., alias='defaultPayCurrency')
    default_pay_source: Optional[int] = Field(None, alias='defaultPaySource')
    default_pay_account_alias: Optional[str] = Field(None, alias='defaultPayAccountAlias')
    email: Optional[str] = None
    first_transaction_id: Union[str, int, None] = Field(None, alias='firstTxnId')
    language: str
    operator: str
    phone_hash: str = Field(alias='phoneHash')
    promo_enabled: Optional[bool] = Field(None, alias='promoEnabled')

    @validator('default_pay_currency', pre=True)
    def humanize_pay_currency(cls, v):  # type: ignore
        if not isinstance(v, int):
            return v
        return Currency.get(str(v))


class UserProfile(Base):
    auth_info: Optional[AuthInfo] = Field(None, alias='authInfo')
    contract_info: Optional[ContractInfo] = Field(None, alias='contractInfo')
    user_info: Optional[UserInfo] = Field(None, alias='userInfo')


__all__ = ['UserProfile']



from typing import Optional

from pydantic import Field

from glQiwiApi.types.base import Base


class NickName(Base):
    can_change: bool = Field(..., alias='canChange')
    can_use: bool = Field(..., alias='canUse')
    description: str
    nickname: Optional[str] = None



from typing import ClassVar, List, Optional

from pydantic import Field

from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import Card


class GetBoundedCards(QiwiAPIMethod[List[Card]]):
    http_method: ClassVar[str] = 'GET'
    url: ClassVar[str] = 'https://edge.qiwi.com/cards/v1/cards'

    card_alias: Optional[str] = Field(None, alias='vas-alias')



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, cast

from glQiwiApi.qiwi.clients.wallet.types.transaction import Transaction
from glQiwiApi.types.amount import AmountWithCurrency

from .base import Handler

if TYPE_CHECKING:
    from glQiwiApi.qiwi.clients.wallet.client import QiwiWallet


class AbstractTransactionHandler(Handler[Transaction], abc.ABC):
    @property
    def wallet(self) -> QiwiWallet:
        return cast(QiwiWallet, self.context['wallet'])

    @property
    def transaction_id(self) -> int:
        return self.event.id

    @property
    def transaction_sum(self) -> AmountWithCurrency:
        return self.event.sum



from pydantic import BaseModel, Field


class Polygon(BaseModel):
    """Polygon class for QiwiMaps class"""

    latitude_north_western: float = Field(..., alias='latNW')
    longitude_north_western: float = Field(..., alias='lngNW')
    latitude_south_east: float = Field(..., alias='latSE')
    longitude_south_east: float = Field(..., alias='lngSE')



import re
from typing import Any, ClassVar, List, cast

from glQiwiApi.core.abc.api_method import APIMethod, Request, ReturningType
from glQiwiApi.core.session.holder import HTTPResponse

YOO_MONEY_LINK_REGEXP = re.compile(r'https://yoomoney.ru/oauth2/authorize[?]requestid[=]\w+')


class BuildAuthURL(APIMethod[str]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://yoomoney.ru/oauth/authorize'

    client_id: str
    scopes: List[str]
    redirect_uri: str
    response_type: str = 'code'

    def build_request(self, **url_format_kw: Any) -> 'Request':
        return Request(
            endpoint=self.url.format(**url_format_kw, **self._get_runtime_path_values()),
            http_method=self.http_method,
            data={
                'client_id': self.client_id,
                'response_type': 'code',
                'redirect_uri': self.redirect_uri,
                'scope': ' '.join(self.scopes),
            },
        )

    @classmethod
    def parse_http_response(cls, response: HTTPResponse) -> ReturningType:
        try:
            return cast(
                str, re.findall(YOO_MONEY_LINK_REGEXP, response.body.decode('utf-8'))[0]
            )  # pragma: no cover
        except IndexError:
            raise Exception(
                'Could not find the authorization link in the response from '
                'the api, check the client_id value'
            )



import logging
from typing import Any, Dict, Optional, TypeVar, cast

from aiohttp.typedefs import LooseCookies

from glQiwiApi.core.abc.api_method import APIMethod
from glQiwiApi.core.cache.cached_types import CachedAPIRequest, Payload
from glQiwiApi.core.cache.storage import CacheStorage
from glQiwiApi.core.session.holder import AbstractSessionHolder, AiohttpSessionHolder, HTTPResponse
from glQiwiApi.utils.compat import Protocol
from glQiwiApi.utils.payload import make_payload

logger = logging.getLogger('glQiwiApi.RequestService')

T = TypeVar('T')


class RequestServiceProto(Protocol):
    async def execute_api_method(self, method: APIMethod[T], **url_kw: Any) -> T:
        ...

    async def get_json_content(
        self,
        url: str,
        method: str,
        cookies: Optional[LooseCookies] = None,
        json: Optional[Any] = None,
        data: Optional[Any] = None,
        headers: Optional[Any] = None,
        params: Optional[Any] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        ...

    async def send_request(
        self,
        url: str,
        method: str,
        cookies: Optional[LooseCookies] = None,
        json: Optional[Any] = None,
        data: Optional[Any] = None,
        headers: Optional[Any] = None,
        params: Optional[Any] = None,
        **kwargs: Any,
    ) -> HTTPResponse:
        ...

    async def warmup(self) -> Any:
        ...

    async def shutdown(self) -> None:
        ...


class RequestService:
    def __init__(
        self,
        session_holder: Optional[AbstractSessionHolder[Any]] = None,
    ) -> None:
        if session_holder is None:
            session_holder = AiohttpSessionHolder()
        self._session_holder = session_holder

    async def execute_api_method(self, method: APIMethod[T], **url_kw: Any) -> T:
        request = method.build_request(**url_kw)
        raw_http_response = await self.send_request(
            request.endpoint,
            request.http_method,
            params=request.params,
            data=request.data,
            headers=request.headers,
            json=request.json_payload,
        )
        return method.parse_http_response(raw_http_response)

    async def get_json_content(
        self,
        url: str,
        method: str,
        cookies: Optional[LooseCookies] = None,
        json: Optional[Any] = None,
        data: Optional[Any] = None,
        headers: Optional[Any] = None,
        params: Optional[Any] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        prepared_payload = make_payload(**locals(), exclude=('kwargs',))
        response = await self.send_request(**prepared_payload)
        return response.json()

    async def warmup(self) -> Any:
        return await self._session_holder.get_session()

    async def shutdown(self) -> None:
        await self._session_holder.close()

    async def send_request(
        self,
        url: str,
        method: str,
        cookies: Optional[LooseCookies] = None,
        json: Optional[Any] = None,
        data: Optional[Any] = None,
        headers: Optional[Any] = None,
        params: Optional[Any] = None,
        **kwargs: Any,
    ) -> HTTPResponse:
        session = await self._session_holder.get_session()
        return await self._session_holder.convert_third_party_lib_response_to_http_response(
            await session.request(
                method=method,
                url=url,
                data=data,
                headers=headers,
                json=json,
                cookies=cookies,
                params=params,
                **kwargs,
            )
        )


class RequestServiceCacheDecorator(RequestServiceProto):
    def __init__(
        self,
        request_service: RequestServiceProto,
        cache_storage: CacheStorage,
    ) -> None:
        self._cache = cache_storage
        self._request_service = request_service

    async def execute_api_method(self, method: APIMethod[T], **url_kw: Any) -> T:
        return await self._request_service.execute_api_method(method, **url_kw)

    async def get_json_content(
        self,
        url: str,
        method: str,
        cookies: Optional[LooseCookies] = None,
        json: Optional[Any] = None,
        data: Optional[Any] = None,
        headers: Optional[Any] = None,
        params: Optional[Any] = None,
        **kwargs: Any,
    ) -> Dict[Any, Any]:
        request_args = make_payload(**locals())

        if await self._cache.contains_similar(Payload(**request_args)):
            return cast(Dict[Any, Any], (await self._cache.retrieve(url)).response)

        response = await self._request_service.get_json_content(**request_args)
        await self._cache_response(response, **request_args)
        return response

    async def send_request(
        self,
        url: str,
        method: str,
        cookies: Optional[LooseCookies] = None,
        json: Optional[Any] = None,
        data: Optional[Any] = None,
        headers: Optional[Any] = None,
        params: Optional[Any] = None,
        **kwargs: Any,
    ) -> HTTPResponse:
        return await self._request_service.send_request(
            url, method, cookies, json, data, headers, params, **kwargs
        )

    async def warmup(self) -> Any:
        return await self._request_service.shutdown()

    async def shutdown(self) -> None:
        await self._request_service.shutdown()

    async def _cache_response(self, response: Any, method: str, **kwargs: Any) -> None:
        await self._cache.update(
            **{
                kwargs['endpoint']: CachedAPIRequest(
                    payload=Payload(**kwargs), response=response, method=method
                )
            }
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['glQiwiApi',
 'glQiwiApi.contrib',
 'glQiwiApi.core',
 'glQiwiApi.core.abc',
 'glQiwiApi.core.cache',
 'glQiwiApi.core.event_fetching',
 'glQiwiApi.core.event_fetching.class_based',
 'glQiwiApi.core.event_fetching.webhooks',
 'glQiwiApi.core.event_fetching.webhooks.dto',
 'glQiwiApi.core.event_fetching.webhooks.middlewares',
 'glQiwiApi.core.event_fetching.webhooks.services',
 'glQiwiApi.core.event_fetching.webhooks.services.security',
 'glQiwiApi.core.event_fetching.webhooks.views',
 'glQiwiApi.core.session',
 'glQiwiApi.ext',
 'glQiwiApi.plugins',
 'glQiwiApi.plugins.aiogram',
 'glQiwiApi.qiwi',
 'glQiwiApi.qiwi.clients',
 'glQiwiApi.qiwi.clients.maps',
 'glQiwiApi.qiwi.clients.maps.methods',
 'glQiwiApi.qiwi.clients.maps.types',
 'glQiwiApi.qiwi.clients.p2p',
 'glQiwiApi.qiwi.clients.p2p.methods',
 'glQiwiApi.qiwi.clients.wallet',
 'glQiwiApi.qiwi.clients.wallet.methods',
 'glQiwiApi.qiwi.clients.wallet.methods.qiwi_master',
 'glQiwiApi.qiwi.clients.wallet.methods.webhook',
 'glQiwiApi.qiwi.clients.wallet.types',
 'glQiwiApi.types',
 'glQiwiApi.types.arbitrary',
 'glQiwiApi.utils',
 'glQiwiApi.utils.synchronous',
 'glQiwiApi.yoo_money',
 'glQiwiApi.yoo_money.methods']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0', 'pydantic>=1.10.2,<2.0.0', 'tzdata>=2022.6,<2023.0']

extras_require = \
{':python_full_version <= "3.7.0"': ['typing_extensions>=4.0.1,<5.0.0'],
 ':python_version < "3.9"': ['backports.zoneinfo>=0.2.1,<0.3.0'],
 'docs': ['Sphinx>=4.3.2,<5.0.0',
          'sphinx-intl>=2.0.1,<3.0.0',
          'sphinx-autobuild>=2021.3.14,<2022.0.0',
          'sphinx-copybutton>=0.5.0,<0.6.0',
          'furo>=2022.2.23,<2023.0.0',
          'sphinx-prompt>=1.5.0,<2.0.0',
          'Sphinx-Substitution-Extensions>=2020.9.30,<2023.0.0',
          'towncrier>=21.9.0,<22.0.0',
          'pygments>=2.4,<3.0',
          'pymdown-extensions>=9.4,<10.0',
          'markdown-include>=0.6,<0.8',
          'sphinxemoji',
          'sphinx-notfound-page'],
 'fast': ['aiofiles>=22.1.0,<23.0.0'],
 'fast:sys_platform == "darwin" or sys_platform == "linux"': ['uvloop>=0.16.0,<0.17.0']}

setup_kwargs = {
    'name': 'glqiwiapi',
    'version': '2.18.1',
    'description': 'The ultrarapid and multifunctional wrapper over QIWI and YooMoney',
    'long_description': '<h2 align="center">\n<img src="https://github.com/GLEF1X/glQiwiApi/blob/dev-2.x/docs/_static/logo.png?raw=true" width="200"></img>\n\n\n[![PyPI version](https://img.shields.io/pypi/v/glQiwiApi.svg)](https://pypi.org/project/glQiwiApi/) [![Code Quality Score](https://api.codiga.io/project/20780/score/svg)](https://frontend.code-inspector.com/public/project/20780/glQiwiApi/dashboard) ![Downloads](https://img.shields.io/pypi/dm/glQiwiApi) ![docs](https://readthedocs.org/projects/pip/badge/?version=latest)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/GLEF1X/glQiwiApi.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/GLEF1X/glQiwiApi/context:python) [![CodeFactor](https://www.codefactor.io/repository/github/glef1x/glqiwiapi/badge)](https://www.codefactor.io/repository/github/glef1x/glqiwiapi)\n![codecov](https://codecov.io/gh/GLEF1X/glQiwiApi/branch/dev-2.x/graph/badge.svg?token=OD538HKV15)\n![CI](https://github.com/GLEF1X/glQiwiApi/actions/workflows/tests.yml/badge.svg) ![mypy](https://img.shields.io/badge/%20type_checker-mypy-%231674b1?style=flat) [![Downloads](https://pepy.tech/badge/glqiwiapi/month)](https://pepy.tech/project/glqiwiapi) [![Downloads](https://pepy.tech/badge/glqiwiapi)](https://pepy.tech/project/glqiwiapi)\n\n<img src="https://github.com/GLEF1X/glQiwiApi/blob/dev-2.x/docs/static/demo.gif"/>\n</h2>\n\n## 🌎Official api resources:\n\n* 🎓 __Documentation: [here](https://glqiwiapi.readthedocs.io/en/latest/)__\n* 🖱️ __Telegram chat: [![Dev-Telegram](https://img.shields.io/badge/Telegram-blue.svg?style=flat-square&logo=telegram)](https://t.me/glQiwiAPIOfficial)__\n\n## Benchmarks\n\n```bash\nhint: smaller is better\nglQiwiApi      90.9925 (1.0)      103.3993 (1.0)       95.4082 (1.0)      5.3941 (1.0)       92.4023 (1.0)       8.2798 (1.0)\npyQiwiP2P     112.2819 (1.23)     135.0227 (1.31)     123.7498 (1.30)     9.9919 (1.85)     127.5926 (1.38)     17.2723 (2.09)\n```\n\n## 🐦Dependencies\n\n| Library |                       Description                       |\n|:-------:|:-------------------------------------------------------:|\n|aiohttp  | Asynchronous HTTP Client/Server for asyncio and Python. |\n|pydantic |    Json data validator. Very fast instead of custom     |\n',
    'author': 'Glib Garanin',
    'author_email': 'glebgar567@gmail.com',
    'maintainer': 'GLEF1X',
    'maintainer_email': 'glebgar567@gmail.com',
    'url': 'https://github.com/GLEF1X/glQiwiApi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import asyncio
import concurrent.futures as futures
import functools as ft
from contextvars import ContextVar


class AdaptiveExecutor(futures.ThreadPoolExecutor):
    """object: AdaptiveExecutor"""

    def __init__(self, max_workers=None, **kwargs):
        super().__init__(max_workers, 'sync_adapter_', **kwargs)
        self.max_workers = max_workers
        self.is_from_running_loop = ContextVar('Adapter_', default=False)


def execute_async_as_sync(func, *args, **kwargs):
    """
    Function to execute async functions synchronously

    :param func: Async function, which you want to execute in synchronous code
    :param args: args, which need your async func
    :param kwargs: kwargs, which need your async func
    """
    try:
        shutdown_callback = kwargs.pop('__shutdown__callback__')
    except KeyError:
        shutdown_callback = None
    loop, executor = _construct_executor_and_loop()
    wrapped_future = asyncio.run_coroutine_threadsafe(func(*args, **kwargs), loop=loop)
    executor.submit(run_forever_safe, loop, shutdown_callback)
    try:
        # Get result
        return await_sync(wrapped_future)
    finally:
        # Cleanup
        _on_shutdown(executor, loop)


def _construct_executor_and_loop():
    """Get or create new event loop"""
    loop = take_event_loop()
    executor = AdaptiveExecutor()
    loop.set_default_executor(executor)
    return loop, executor


def take_event_loop(set_debug: bool = False):
    """
    Get new or running event loop

    :param set_debug:
    """
    try:
        loop = asyncio.get_running_loop()
        if loop.is_closed():
            raise RuntimeError()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    loop.set_debug(set_debug)
    return loop


def run_forever_safe(loop, callback=None) -> None:
    """run a loop for ever and clean up after being stopped"""

    loop.run_forever()
    # NOTE: loop.run_forever returns after calling loop.stop

    safe_cancel(loop=loop, callback=callback)


def safe_cancel(loop, callback) -> None:
    """cancel all tasks and close the loop gracefully"""

    loop_tasks_all = asyncio.all_tasks(loop=loop)

    # execute shutdown callback to gracefully shutdown your components
    if callback is not None:
        loop.run_until_complete(callback())

    # NOTE: `cancel` does not guarantee that the task will be cancelled
    for task in loop_tasks_all:
        task.cancel()

    for task in loop_tasks_all:
        if not (task.done() or task.cancelled()):
            try:
                # wait for task cancellations
                loop.run_until_complete(task)
            except asyncio.CancelledError:
                pass
    # Finally, close event loop
    loop.close()


def _cancel_future(loop, future, executor) -> None:
    """cancels future if any exception occurred"""
    executor.submit(loop.call_soon_threadsafe, future.cancel)


def _stop_loop(loop) -> None:
    """stops an event loop"""
    loop.stop()


def await_sync(future):
    """synchronously waits for a task"""
    return future.result()


def _on_shutdown(executor, loop):
    """Do some cleanup"""
    if not executor.is_from_running_loop.get():
        loop.call_soon_threadsafe(_stop_loop, loop)
    executor.shutdown(wait=True)


class async_as_sync:  # NOQA
    def __init__(self, async_shutdown_callback=None, sync_shutdown_callback=None):
        self._sync_shutdown_callback = sync_shutdown_callback
        self._async_shutdown_callback = async_shutdown_callback

    def __call__(self, func):
        @ft.wraps(func)
        def wrapper(*args, **kwargs):
            result = execute_async_as_sync(
                func, *args, __shutdown__callback__=self._async_shutdown_callback, **kwargs
            )
            if self._sync_shutdown_callback is not None:
                return self.execute_sync_callback(result)
            return result

        return wrapper

    def execute_sync_callback(self, result):
        return self._sync_shutdown_callback(result)



from datetime import datetime, timedelta
from typing import Any, ClassVar, Dict, List, Optional

from pydantic import Field, root_validator

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import Statistic, TransactionType
from glQiwiApi.utils.date_conversion import datetime_to_iso8601_with_moscow_timezone


class FetchStatistics(QiwiAPIMethod[Statistic]):
    url: ClassVar[
        str
    ] = 'https://edge.qiwi.com/payment-history/v2/persons/{phone_number}/payments/total'
    http_method: ClassVar[str] = 'GET'

    start_date: datetime = Field(default_factory=datetime.now)
    end_date: datetime = Field(default_factory=lambda: datetime.now() - timedelta(days=90))
    operation: TransactionType = TransactionType.ALL
    sources: Optional[List[str]] = None

    @root_validator()
    def check_start_date_and_end_date_difference_not_greater_than_90_days(
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        start_date: Optional[datetime] = values.get('start_date')
        end_date: Optional[datetime] = values.get('end_date')

        if start_date is None or end_date is None:  # denotes that type of arguments is invalid
            return values

        if (end_date - start_date).days > 90 or (start_date - end_date).days > 90:
            raise ValueError('The maximum period for downloading statistics is 90 calendar days.')

        return values

    def build_request(self, **url_format_kw: Any) -> Request:
        params = {
            'startDate': datetime_to_iso8601_with_moscow_timezone(self.start_date),
            'endDate': datetime_to_iso8601_with_moscow_timezone(self.end_date),
            'operation': self.operation.value,
        }
        if self.sources:
            params.update({'sources': ' '.join(self.sources)})
        return Request(
            endpoint=self.url.format(**url_format_kw), params=params, http_method=self.http_method
        )



from .base import Handler
from .bill import AbstractBillHandler
from .error import ErrorHandler
from .transaction import AbstractTransactionHandler
from .webhook_transaction import AbstractTransactionWebhookHandler

__all__ = (
    'AbstractTransactionHandler',
    'AbstractBillHandler',
    'Handler',
    'AbstractTransactionWebhookHandler',
    'ErrorHandler',
)



from typing import Optional

from pydantic import Field

from glQiwiApi.types.base import Base


class Coordinate(Base):
    """Object: coordinate"""

    latitude: float = Field(..., alias='latitude')
    longitude: float = Field(..., alias='longitude')
    precision: int = Field(..., alias='precision')


class Terminal(Base):
    """Object: Terminal"""

    terminal_id: int = Field(..., alias='terminalId')
    ttp_id: int = Field(..., alias='ttpId')
    last_active: str = Field(..., alias='lastActive')
    count: int = Field(..., alias='count')
    address: str = Field(..., alias='address')
    verified: bool = Field(..., alias='verified')
    label: str = Field(..., alias='label')
    description: Optional[str] = Field(type(None), alias='description')
    cash_allowed: bool = Field(..., alias='cashAllowed')
    card_allowed: bool = Field(..., alias='cardAllowed')
    identification_type: int = Field(..., alias='identificationType')
    coordinate: Coordinate = Field(..., alias='coordinate')



import abc
from typing import Any, Generic, Set, TypeVar

T = TypeVar('T')


class UnexpectedCollision(Exception):
    pass


class UnhashableObjectError(TypeError):
    pass


class AbstractCollisionDetector(abc.ABC, Generic[T]):
    """
    QIWI API can transfer_money the same update twice or more, so we need to avoid this problem anyway.
    Also, you can override it with redis usage or more advanced hashing.
    """

    @abc.abstractmethod
    def has_collision(self, obj: T) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def add_already_processed_event(self, obj: T) -> None:
        raise NotImplementedError

    def remember_processed_object(self, obj: T) -> None:
        if self.has_collision(obj):
            raise UnexpectedCollision()
        self.add_already_processed_event(obj)


class HashBasedCollisionDetector(AbstractCollisionDetector[T]):
    def __init__(self) -> None:
        self.already_processed_object_hashes: Set[int] = set()

    def add_already_processed_event(self, obj: T) -> None:
        if _is_object_unhashable(obj):
            raise UnhashableObjectError(f'Object {obj!r} is unhashable')
        self.already_processed_object_hashes.add(hash(obj))

    def has_collision(self, obj: T) -> bool:
        if _is_object_unhashable(obj):
            raise UnhashableObjectError(f'Object {obj!r} is unhashable')
        return any(
            hash(obj) == processed_hash for processed_hash in self.already_processed_object_hashes
        )


def _is_object_unhashable(obj: Any) -> bool:
    try:
        hash(obj)
        return False
    except TypeError:
        return True



from typing import Optional, Union

from pydantic import BaseConfig, BaseModel, Field, validator

from glQiwiApi.types.base import Base, HashableBase


class CurrencyModel(HashableBase):
    code: str
    decimal_digits: int
    name: str
    name_plural: str
    rounding: Union[int, float]
    symbol: str
    symbol_native: str
    iso_format: Optional[str] = Field(..., alias='isoformat')

    def __str__(self) -> str:
        return self.code

    class Config(BaseConfig):
        frozen = True
        allow_mutation = False


class AmountWithCurrency(Base):
    amount: float
    currency: Union[CurrencyModel, str]  # string if currency util couldn't parse it

    @validator('currency', pre=True)
    def humanize_pay_currency(cls, v):  # type: ignore
        from glQiwiApi.utils.currency_util import Currency

        if not isinstance(v, int):
            try:
                v = int(v)
            except ValueError:
                return v
        return Currency.get(str(v))


class HashableSum(HashableBase, AmountWithCurrency):
    ...


class PlainAmount(BaseModel):
    value: float
    currency: str


class HashablePlainAmount(HashableBase, PlainAmount):
    ...


class Type(BaseModel):
    id: str
    title: str



from datetime import date
from typing import Optional

from pydantic import Field

from glQiwiApi.types.base import Base


class Identification(Base):
    """object: Identification"""

    identification_id: int = Field(..., alias='id')
    first_name: str = Field(..., alias='firstName')
    middle_name: str = Field(..., alias='middleName')
    last_name: str = Field(..., alias='lastName')
    birth_date: date = Field(..., alias='birthDate')
    passport: str
    inn: Optional[str]
    snils: Optional[str]
    oms: Optional[str]
    type: str



from datetime import datetime
from typing import Union

from pydantic import Field, validator

from glQiwiApi.types.amount import CurrencyModel
from glQiwiApi.types.base import Base
from glQiwiApi.utils.currency_util import Currency


class Interval(Base):
    """object: Interval"""

    date_from: datetime = Field(alias='dateFrom')
    date_till: datetime = Field(alias='dateTill')


class Limit(Base):
    """object: Limit"""

    currency: CurrencyModel
    rest: Union[float, int]
    max_limit: Union[float, int] = Field(alias='max')
    spent: Union[float, int]
    interval: Interval
    limit_type: str = Field(alias='type')

    @validator('currency', pre=True)
    def currency_validate(cls, v):  # type: ignore
        if not isinstance(v, str):
            raise ValueError()
        return Currency.get(v)


__all__ = ['Limit']



from typing import Any, ClassVar, Dict

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import OrderDetails


class CreateCardPurchaseOrder(QiwiAPIMethod[OrderDetails]):
    url: ClassVar[str] = 'https://edge.qiwi.com/cards/v2/persons/{phone_number}/orders'
    http_method: ClassVar[str] = 'POST'

    json_payload_schema: ClassVar[Dict[str, Any]] = {'cardAlias': RuntimeValue()}

    card_alias: str = Field(..., schema_path='cardAlias')



import abc
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseConfig, BaseModel, Extra, parse_obj_as
from pydantic.fields import ModelField
from pydantic.generics import GenericModel

from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.utils.compat import json

ReturningType = TypeVar('ReturningType')
_T = TypeVar('_T')
_sentinel = object()


class RuntimeValueIsMissing(Exception):
    pass


def _filter_none_values(d: Dict[str, Any]) -> Dict[str, Any]:
    return {k: v for k, v in d.items() if v}


DEFAULT_EXCLUDE = {'request_schema', 'endpoint', 'http_method'}


class APIMethod(abc.ABC, GenericModel, Generic[ReturningType]):
    """
    Reflect the real API method that could be emitted using appropriate clients.
    """

    class Config(BaseConfig):
        extra = Extra.allow
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        orm_mode = True
        underscore_attrs_are_private = True
        json_loads = json.loads
        json_dumps = json.dumps  # type: ignore

    __returning_type__: ClassVar[Any] = _sentinel

    json_payload_schema: ClassVar[Dict[str, Any]] = {}

    @property
    @abc.abstractmethod
    def url(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def http_method(self) -> str:
        pass

    def __class_getitem__(cls, params: Union[Type[Any], Tuple[Type[Any], ...]]) -> Type[Any]:
        """
        Allows us to get generic class in runtime instead of do it explicitly
        in every class that inherits from APIMethod.

        @param params:
        @return:
        """
        if isinstance(params, tuple):
            return super().__class_getitem__(params)

        if not params and cls is not Tuple:
            raise TypeError(f'Parameter list to {cls.__qualname__}[...] cannot be empty')

        key = params  # just alias

        if cls.__returning_type__ is _sentinel or cls.__returning_type__ is ReturningType:  # type: ignore
            cls.__returning_type__ = key
        else:
            try:
                cls.__returning_type__ = cls.__returning_type__[key]
            except TypeError:
                cls.__returning_type__ = key

        return super().__class_getitem__(params)

    def __init_subclass__(cls) -> None:
        if cls.__returning_type__ is not _sentinel:
            cls.__returning_type__ = cls.__returning_type__

    @classmethod
    def parse_http_response(cls, response: HTTPResponse) -> ReturningType:
        if cls.__returning_type__ is _sentinel or cls.__returning_type__ is ReturningType:  # type: ignore
            raise RuntimeError(f'{cls.__qualname__}: __returning_type__ is missing')

        json_response = response.json()

        manually_parsed_json = cls.on_json_parse(response)
        if manually_parsed_json is not _sentinel:
            return manually_parsed_json

        try:
            if issubclass(cls.__returning_type__, BaseModel):
                return cast(ReturningType, cls.__returning_type__.parse_obj(json_response))
        except TypeError:  # issubclass() arg 1 must be a class
            pass

        return cast(ReturningType, parse_obj_as(cls.__returning_type__, json_response))

    @classmethod
    def on_json_parse(cls, response: HTTPResponse) -> Union[Any, ReturningType]:
        return _sentinel

    def build_request(self, **url_format_kw: Any) -> 'Request':
        request_kw: Dict[str, Any] = {
            'endpoint': self.url.format(**url_format_kw, **self._get_runtime_path_values()),
            'http_method': self.http_method,
        }

        if self.http_method == 'GET' and self.json_payload_schema:
            raise TypeError(
                'Request schema is not compatible with GET http method '
                'because GET method cannot transfer json payload'
            )

        if self.http_method == 'GET':
            request_kw['params'] = self.dict(
                exclude_none=True, by_alias=True, exclude=self._get_exclude_set()
            )

        if self.json_payload_schema:
            request_kw['json_payload'] = self._get_filled_json_payload_schema()
        else:
            request_kw['data'] = self.dict(
                exclude_none=True, by_alias=True, exclude=self._get_exclude_set()
            )

        return Request(**_filter_none_values(request_kw))

    def _get_exclude_set(self) -> Set[str]:
        to_exclude: Set[str] = DEFAULT_EXCLUDE.copy()
        for key, field in self.__fields__.items():
            if field.field_info.extra.get('path_runtime_value', False):
                to_exclude.add(key)

        return to_exclude

    def _get_filled_json_payload_schema(self) -> Dict[str, Any]:
        """
        Algorithm of this function firstly takes care of default values,
        if field has no default value for schema it checks values, that were transmitted to method

        @return:
        """
        request_schema = self._get_schema_with_filled_runtime_values()
        schema_values = self.dict(
            exclude_none=True, by_alias=True, exclude=self._get_exclude_set()
        )

        for key, field in self.__fields__.items():
            scheme_path: str = field.field_info.extra.get('scheme_path', field.name)
            keychain = scheme_path.split('.')
            try:
                field_value = schema_values[key]
            except KeyError:
                continue
            _insert_value_into_dictionary(request_schema, keychain, field_value)

        return request_schema

    def _get_schema_with_filled_runtime_values(self) -> Dict[str, Any]:
        scheme_paths: List[str] = [
            field.field_info.extra.get('scheme_path', field.name)
            for field in self.__fields__.values()
            if field.name in self.dict(exclude_none=True)
        ]
        schema = self.json_payload_schema.copy()

        def apply_runtime_values_to_schema(d: Dict[str, Any]) -> None:
            for k, v in list(d.items()):
                if isinstance(v, dict):
                    apply_runtime_values_to_schema(v)
                elif isinstance(v, RuntimeValue):
                    if v.has_default() and k not in scheme_paths:
                        d[k] = v.get_default()
                        continue
                    elif v.is_mandatory is False:
                        d.pop(k)

        apply_runtime_values_to_schema(schema)

        return schema

    def _get_runtime_path_values(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {}

        for field_name, value in self.dict().items():
            field: ModelField = self.__fields__[field_name]
            is_path_runtime_value = field.field_info.extra.get('path_runtime_value', False)
            if not is_path_runtime_value:
                continue

            result[field_name] = value

        return result


class Request(BaseModel):
    endpoint: str

    data: Optional[Dict[str, Optional[Any]]] = None
    params: Optional[Dict[str, Optional[Any]]] = None
    json_payload: Optional[Dict[str, Any]] = None
    headers: Dict[str, Any] = {}

    http_method: str = 'GET'

    class Config(BaseConfig):
        arbitrary_types_allowed = True


class RuntimeValue:
    __slots__ = ('_default', '_default_factory', 'is_mandatory')

    def __init__(
        self,
        default: Optional[Any] = None,
        default_factory: Optional[Callable[..., Any]] = None,
        mandatory: bool = True,
    ):
        self._default = default
        self._default_factory = default_factory
        self.is_mandatory = mandatory

    def has_default(self) -> bool:
        return self._default is not None or self._default_factory is not None

    def get_default(self) -> Any:
        if self._default is not None:
            return self._default
        if self._default_factory is not None:
            return self._default_factory()


def _insert_value_into_dictionary(
    d: Dict[str, _T], keychain: List[str], value: Any
) -> None:  # pragma: no cover
    if not keychain:
        return None

    if len(keychain) == 1:
        d[keychain[0]] = value

    for k, v in d.items():
        if not isinstance(v, dict):
            continue
        next_key_from_keychain = keychain[0]

        if k != next_key_from_keychain:
            continue

        _insert_value_into_dictionary(v, keychain[1:], value)



from typing import ClassVar, List

from pydantic import parse_obj_as

from glQiwiApi.core.abc.api_method import ReturningType
from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import Balance


class GetBalances(QiwiAPIMethod[List[Balance]]):
    http_method: ClassVar[str] = 'GET'
    url: ClassVar[str] = 'https://edge.qiwi.com/funding-sources/v2/persons/{phone_number}/accounts'

    @classmethod
    def on_json_parse(cls, response: HTTPResponse) -> List[Balance]:
        return parse_obj_as(List[Balance], response.json()['accounts'])



import base64
import enum
import hashlib
import hmac
from datetime import datetime
from typing import Any, Dict, List, Optional, cast

from pydantic import Field, root_validator

from glQiwiApi.types.amount import HashableSum
from glQiwiApi.types.base import Base, HashableBase
from glQiwiApi.types.exceptions import WebhookSignatureUnverifiedError


class WebhookPayment(HashableBase):
    """Scheme of webhook payment object"""

    account: str = Field(..., alias='account')
    comment: str = Field(..., alias='comment')
    date: datetime = Field(..., alias='date')
    error_code: str = Field(..., alias='errorCode')
    person_id: int = Field(..., alias='personId')
    provider: int = Field(..., alias='provider')
    sign_fields: str = Field(..., alias='signFields')
    status: str = Field(..., alias='status')
    txn_id: str = Field(..., alias='txnId')
    type: str = Field(..., alias='type')
    commission: Optional[HashableSum] = Field(default=None, alias='calc_commission')
    sum: HashableSum = Field(..., alias='sum')
    total: HashableSum = Field(..., alias='total')


class TransactionWebhook(HashableBase):
    """Object: TransactionWebhook"""

    hash: Optional[str] = Field(default=None, alias='hash')
    id: str = Field(..., alias='hookId')
    message_id: Optional[str] = Field(default=None, alias='messageId')
    is_experimental: bool = Field(..., alias='test')
    version: str = Field(..., alias='version')
    payment: Optional[WebhookPayment] = Field(default=None, alias='payment')

    signature: Optional[str] = None
    """
    NOT API field, it's generating by method `_webhook_signature_collector`
    Don't rely on it, if you want to use signature, generate new one using the same logic as in validator
    """

    def verify_signature(self, webhook_base64_key: str) -> None:
        if self.signature is None:
            raise WebhookSignatureUnverifiedError('Signature attribute is None')

        webhook_key = base64.b64decode(bytes(webhook_base64_key, 'utf-8'))
        generated_hash = hmac.new(
            webhook_key, self.signature.encode('utf-8'), hashlib.sha256
        ).hexdigest()
        if generated_hash != self.hash:
            raise WebhookSignatureUnverifiedError()

    @root_validator(pre=True)
    def _webhook_signature_collector(cls, values: Dict[Any, Any]) -> Dict[Any, Any]:
        """
        Get webhook signature to confirm it with hash by base64 encoded key.
        payment.signFields is string. e.g. "sum.currency,sum.amount,type,account,txnId".
        So, this string disassembled piece by piece recursively, because each of signField part, that separated by comma
        can be nested by "."

        @param values:
        @return:
        """
        payment = cast(Optional[Dict[Any, Any]], values.get('payment'))
        if payment is None:
            return values

        sign_fields = payment.get('signFields')
        if sign_fields is None:
            raise ValueError('Cannot generate signature, sign fields is empty')

        sign_fields_list = cast(str, sign_fields).split(',')
        webhook_signature = '|'.join(
            str(_get_sign_field(payment, sign_field.split('.'))) for sign_field in sign_fields_list
        )
        values.update(signature=webhook_signature)
        return values


def _get_sign_field(dictionary: Dict[Any, Any], nested_keys_list: List[str]) -> Any:
    """
    Recursively iter for nested_keys_list and get nested element.

    For example, imagine that we have list(nested_keys_list) like ["x", "y", "z"],
    x and y is a dictionaries and z is our desired value. So, it's not trivial task without recursion.

    Also, imagine that our nested dict(dictionary) is {"x": {"y": "some_value"}} or {"x": {"y": {"z": "some_value"}}}.
    So, in first example z is already value, in second z is a dict with the desired value and with this help function we
    can process both of this cases.

    @param dictionary:
    @param nested_keys_list:
    @return:
    """
    current = dictionary.get(nested_keys_list.pop(0))
    if isinstance(current, dict):
        return _get_sign_field(current, nested_keys_list)
    else:
        return current


class HookParameters(Base):
    url: str


class WebhookTransactionType(str, enum.Enum):
    IN = 'IN'
    OUT = 'OUT'
    BOTH = 'BOTH'


class WebhookInfo(Base):
    id: str = Field(..., alias='hookId')
    type: str = Field(..., alias='hookType')
    txn_type: WebhookTransactionType = Field(..., alias='txnType')
    hook_parameters: HookParameters = Field(..., alias='hookParameters')


__all__ = ('WebhookInfo', 'TransactionWebhook', 'WebhookPayment')



import functools
import inspect
import sys
from asyncio import Transport
from typing import Any, Mapping, Tuple, Type, TypeVar, cast, no_type_check

from aiohttp import web
from aiohttp.abc import AbstractView
from aiohttp.web_request import Request

from glQiwiApi.core.event_fetching.webhooks.services.security.ip import IPFilter


def check_ip(ip_filter: IPFilter, request: web.Request) -> Tuple[str, bool]:
    # Try to resolve client IP over reverse proxy
    forwarded_for = request.headers.get('X-Forwarded-For', '')
    if forwarded_for:
        forwarded_for, *_ = forwarded_for.split(',', maxsplit=1)
        return forwarded_for, forwarded_for in ip_filter

    peer_name = cast(Transport, request.transport).get_extra_info('peername')
    # When reverse proxy is not configured IP address can be resolved from incoming connection
    if peer_name:
        host, _ = peer_name
        return host, host in ip_filter

    # Potentially impossible case
    return '', False  # pragma: no cover


View = TypeVar('View', bound=Type[AbstractView])


def inject_dependencies(view: View, dependencies: Mapping[str, Any]) -> View:
    params = inspect.signature(view.__init__).parameters

    deps = {
        name: dependency
        for name, dependency in dependencies.items()
        if not isinstance(dependency, (Request, AbstractView)) and name in params
    }

    return cast(View, partial_class(view.__qualname__, view, **deps))  # type: ignore


@no_type_check
def partial_class(name, cls, *args, **kwds):
    new_cls = type(
        name, (cls,), {'__init__': functools.partialmethod(cls.__init__, *args, **kwds)}
    )

    # The following is copied nearly ad verbatim from `namedtuple's` source.
    """
    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in enviroments where
    # sys._getframe is not defined (Jython for example) or sys._getframe is not
    # defined for arguments greater than 0 (IronPython).
    """
    try:
        new_cls.__module__ = sys._getframe(1).f_globals.get('__name__', '__main__')  # noqa
    except (AttributeError, ValueError):  # pragma: no cover
        pass

    return new_cls



from typing import ClassVar

from pydantic import Field

from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.p2p.types import Bill


class GetBillByID(QiwiAPIMethod[Bill]):
    http_method: ClassVar[str] = 'GET'
    url: ClassVar[str] = 'https://api.qiwi.com/partner/bill/v1/bills/{bill_id}'

    bill_id: str = Field(..., path_runtime_value=True)



from json import JSONDecodeError
from typing import Any, ClassVar, Dict, List, NoReturn, Optional, Union, cast

try:
    from orjson import JSONDecodeError as OrjsonJSONDecodeError
except ImportError:
    OrjsonJSONDecodeError = JSONDecodeError  # noqa  # type: ignore

from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.utils.compat import json

HTTP_STATUS_MATCH_TO_ERROR = {
    400: 'Query syntax error (invalid data format). Can be related to wrong arguments,'
    ' that you have passed to method',
    401: 'Wrong API token or token expired',
    403: 'No permission for this request(API token has insufficient permissions)',
    404: 'Object was not found or there are no objects with the specified characteristics',
    423: 'Too many requests, the service is temporarily unavailable',
    422: 'The domain / subnet / host is incorrectly specified'
    'webhook (in the new_url parameter for the webhook URL),'
    'the hook type or transaction type is incorrectly specified,'
    'an attempt to create a hook if there is one already created',
    405: 'Error related to the type of API request, contact the developer or open an issue',
    500: 'Internal service error',
}


class QiwiAPIError(Exception):
    __slots__ = 'http_response', '_custom_message', '_deserialize_cache', 'error_code', 'message'

    description_en: ClassVar[Optional[str]] = None
    description_ru: ClassVar[Optional[str]] = None

    _error_code_match: ClassVar[Optional[Union[str, int, List[Union[str, int]]]]] = None
    _error_code_contains: ClassVar[Union[str, List[str], None]] = None

    def __init__(self, http_response: HTTPResponse, custom_message: Optional[str] = None):
        self._deserialize_cache: Dict[int, Any] = {}
        self._custom_message = custom_message
        self.http_response = http_response
        self.error_code = self._scaffold_error_code()
        self.message = self._scaffold_error_message()
        self.service_name = self._scaffold_service_name()

    def json(self) -> Dict[str, Any]:
        return self._deserialize_response()

    def raise_exception_matching_error_code(self) -> NoReturn:
        if self.error_code is None:
            raise self

        for ex in self.__class__.__subclasses__():
            err_code_matches: List[Union[int, str]] = ex._error_code_match or []
            if not isinstance(ex._error_code_match, list):
                err_code_matches = [ex._error_code_match]

            err_codes_contains: List[Union[str, int]] = ex._error_code_contains or []
            if not isinstance(err_codes_contains, list):
                err_codes_contains = [err_codes_contains]

            err_code_matches = [str(c) for c in err_code_matches]

            if self.error_code in err_code_matches:
                raise ex(self.http_response, self._custom_message)
            elif any(c in self.error_code.lower() for c in err_codes_contains):
                raise ex(self.http_response, self._custom_message)

        raise self

    def __str__(self) -> str:
        representation = (
            '{message}\n'
            '    * {status_code} HTTP status code\n'
            '    * raw response {raw_response}'
        )
        deserialized_response = self._deserialize_response()

        try:
            raw_response: str = json.dumps(deserialized_response, indent=4, ensure_ascii=False)  # type: ignore
        except Exception:
            raw_response = self.http_response.body.decode('utf-8')

        formatted_representation = representation.format(
            status_code=self.http_response.status_code,
            message=self._compose_error_message(),
            raw_response=raw_response,
        )
        if deserialized_response.get('errorCode') is not None:
            formatted_representation += (
                f"\n    * error code = {deserialized_response['errorCode']}"
            )

        return formatted_representation

    def _compose_error_message(self) -> str:
        """
        Composing error message(-s) executes in two steps

        1) Search error message in dictionary with predefined error messages
        2) Deserialize response and try to know whether there is errorCode,
         that can be interpreted as error message

        If there are two messages founded, then it will be concatenated
        """
        if self._custom_message or self.description_ru or self.description_en:
            return self._custom_message or self.description_ru or self.description_en  # type: ignore
        return HTTP_STATUS_MATCH_TO_ERROR.get(self.http_response.status_code, '')

    def _scaffold_error_code(self) -> Optional[str]:
        r = self._deserialize_response()
        err_code = r.get('errorCode')
        if isinstance(err_code, str):
            return err_code

        err_code = r.get('code')
        if isinstance(err_code, str):
            return err_code.replace('QWPRC-', '')  # type: ignore

        return None

    def _scaffold_error_message(self) -> Optional[str]:
        r = self._deserialize_response()
        return r.get('message') or r.get('description')

    def _scaffold_service_name(self) -> Optional[str]:
        r = self._deserialize_response()
        return r.get('serviceName')

    def _deserialize_response(self) -> Dict[str, Any]:
        """
        Deserializing of response many times is practically zero-cost call,
        because response is being cached after the first deserializing
        """
        if self._deserialize_cache.get(id(self.http_response)) is not None:
            return cast(Dict[str, Any], self._deserialize_cache[id(self.http_response)])

        try:
            content: Dict[str, Any] = json.loads(self.http_response.body)
        except (JSONDecodeError, TypeError, OrjsonJSONDecodeError):
            content = {}

        self._deserialize_cache[id(self.http_response)] = content
        return content


class InternalQIWIError(QiwiAPIError):
    _error_code_match = [3, 749, 750]
    description_en = 'Technical error. Repeat the request later'
    description_ru = 'Техническая ошибка. Повторите платеж позже.'


class IncorrectDataFormat(QiwiAPIError):
    _error_code_match = 4
    description_en = 'Incorrect format of phone or account number. Check the data'
    description_ru = 'Некорректный формат телефона или счета. Проверьте данные.'


class NoSuchNumber(QiwiAPIError):
    _error_code_match = 5
    description_en = 'No such number. Check the data and try again'
    description_ru = 'Данного номера не существует. Проверьте данные и попробуйте еще раз.'


class BankSideReceiptError(QiwiAPIError):
    _error_code_match = 8
    description_en = "Technical problem on the recipient's bank side. Try again later"
    description_ru = 'Техническая проблема на стороне банка-получателя. Попробуйте позже.'


class PaymentUnavailableInYourCountry(QiwiAPIError):
    _error_code_match = 131
    description_en = 'Payment type unavailable for your country'
    description_ru = 'Платеж недоступен для вашей страны'


class NotEnoughFundsError(QiwiAPIError):
    _error_code_match = [220, 407]
    description_en = 'Not enough funds. Replenish your wallet'
    description_ru = 'Недостаточно средств. Пополните кошелек'


class PaymentRejected(QiwiAPIError):
    _error_code_match = [7000, 7600]
    description_en = "Payment rejected. Check card's details and repeat the payment or try to contact the bank that issued the card"
    description_ru = 'Платеж отклонен. Проверьте реквизиты карты и повторите платеж или обратитесь в банк, выпустивший карту'


class ValidationError(QiwiAPIError):
    _error_code_match = [303, 254, 241, 'validation.error', 558, 'internal.invoicing.error']


class ObjectNotFoundError(QiwiAPIError):
    _error_code_contains = 'not.found'


class ReceiptNotAvailable(QiwiAPIError):
    _error_code_contains = 'cheque.not.available'
    description_en = (
        'It is impossible to receive a check due to the fact that '
        'the transaction for this ID has not been completed,'
        'that is, an error occurred during the transaction'
    )


class OperationLimitExceededError(QiwiAPIError):
    _error_code_match = [705, 704, 700, 716, 717]


class ObjectAlreadyExistsError(QiwiAPIError):
    _error_code_contains = ['already.exists']
    description_ru = 'Объект, который вы хотите создать уже был создан ранее'
    description_en = 'Object that you want to create was have already been created earlier'


class MobileOperatorCannotBeDeterminedError(QiwiAPIError):
    pass


class InsufficientTokenRightsError(QiwiAPIError):
    _error_code_match = [309]
    description_ru = 'Недостаточно прав для выполнения данного действия. Чтобы решить эту проблему вам нужно перевыпустить токен с достаточными правами для выполнения нужного вам API метода на сайте QIWI.'
    description_en = 'There are insufficient rights to execute this API method. In order to solve this problem you should regenerate API token with sufficient rights.'
